from google.protobuf import wrappers_pb2 as _wrappers_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class DeviceState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DEVICE_STATE_UNSPECIFIED: _ClassVar[DeviceState]
    DEVICE_STATE_NON_OPERATIONAL: _ClassVar[DeviceState]
    DEVICE_STATE_BOOTED: _ClassVar[DeviceState]
    DEVICE_STATE_CONFIGURED: _ClassVar[DeviceState]
    DEVICE_STATE_CLOCK_READY: _ClassVar[DeviceState]
    DEVICE_STATE_INITIALIZED: _ClassVar[DeviceState]
    DEVICE_STATE_OPERATIONAL: _ClassVar[DeviceState]
    DEVICE_STATE_ERRORED: _ClassVar[DeviceState]

class DeviceError(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DEVICE_ERROR_UNSPECIFIED: _ClassVar[DeviceError]
    DEVICE_ERROR_PLL: _ClassVar[DeviceError]
    DEVICE_ERROR_MOTHERBOARD_FW_UPD: _ClassVar[DeviceError]
    DEVICE_ERROR_MODULES_DETECT: _ClassVar[DeviceError]
    DEVICE_ERROR_SYNTH_FW_UPD: _ClassVar[DeviceError]
    DEVICE_ERROR_SYNTH_COMM: _ClassVar[DeviceError]
    DEVICE_ERROR_START_FPGA: _ClassVar[DeviceError]
    DEVICE_ERROR_ATTACH_FPGA: _ClassVar[DeviceError]
    DEVICE_ERROR_MAIN_PIC: _ClassVar[DeviceError]
    DEVICE_ERROR_GPIO_COMM: _ClassVar[DeviceError]
    DEVICE_ERROR_FPGA_OVERHEAT: _ClassVar[DeviceError]
    DEVICE_ERROR_FPGA_HW_FAIL: _ClassVar[DeviceError]
    DEVICE_ERROR_BAD_INPUT_CLOCK: _ClassVar[DeviceError]
    DEVICE_ERROR_SYNTH_OVERHEAT: _ClassVar[DeviceError]
    DEVICE_ERROR_SYNTH_BAD_READ: _ClassVar[DeviceError]
    DEVICE_ERROR_IF_DOWN_CONV_OVERHEAT: _ClassVar[DeviceError]
    DEVICE_ERROR_IF_DOWN_CONV_BAD_READ: _ClassVar[DeviceError]
    DEVICE_ERROR_RF_DOWN_CONV_OVERHEAT: _ClassVar[DeviceError]
    DEVICE_ERROR_RF_DOWN_CONV_BAD_READ: _ClassVar[DeviceError]
    DEVICE_ERROR_RF_UP_CONV_OVERHEAT: _ClassVar[DeviceError]
    DEVICE_ERROR_RF_UP_CONV_BAD_READ: _ClassVar[DeviceError]

class OctaveModule(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    OCTAVE_MODULE_UNSPECIFIED: _ClassVar[OctaveModule]
    OCTAVE_MODULE_RF_UPCONVERTER: _ClassVar[OctaveModule]
    OCTAVE_MODULE_RF_DOWNCONVERTER: _ClassVar[OctaveModule]
    OCTAVE_MODULE_IF_DOWNCONVERTER: _ClassVar[OctaveModule]
    OCTAVE_MODULE_SYNTHESIZER: _ClassVar[OctaveModule]
    OCTAVE_MODULE_MOTHERBOARD: _ClassVar[OctaveModule]
    OCTAVE_MODULE_SOM: _ClassVar[OctaveModule]

class ConstantSource(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    CONSTANT_SOURCE_UNSPECIFIED: _ClassVar[ConstantSource]
    CONSTANT_SOURCE_50_OHM: _ClassVar[ConstantSource]
    CONSTANT_SOURCE_SHORT: _ClassVar[ConstantSource]
    CONSTANT_SOURCE_OPEN: _ClassVar[ConstantSource]
DEVICE_STATE_UNSPECIFIED: DeviceState
DEVICE_STATE_NON_OPERATIONAL: DeviceState
DEVICE_STATE_BOOTED: DeviceState
DEVICE_STATE_CONFIGURED: DeviceState
DEVICE_STATE_CLOCK_READY: DeviceState
DEVICE_STATE_INITIALIZED: DeviceState
DEVICE_STATE_OPERATIONAL: DeviceState
DEVICE_STATE_ERRORED: DeviceState
DEVICE_ERROR_UNSPECIFIED: DeviceError
DEVICE_ERROR_PLL: DeviceError
DEVICE_ERROR_MOTHERBOARD_FW_UPD: DeviceError
DEVICE_ERROR_MODULES_DETECT: DeviceError
DEVICE_ERROR_SYNTH_FW_UPD: DeviceError
DEVICE_ERROR_SYNTH_COMM: DeviceError
DEVICE_ERROR_START_FPGA: DeviceError
DEVICE_ERROR_ATTACH_FPGA: DeviceError
DEVICE_ERROR_MAIN_PIC: DeviceError
DEVICE_ERROR_GPIO_COMM: DeviceError
DEVICE_ERROR_FPGA_OVERHEAT: DeviceError
DEVICE_ERROR_FPGA_HW_FAIL: DeviceError
DEVICE_ERROR_BAD_INPUT_CLOCK: DeviceError
DEVICE_ERROR_SYNTH_OVERHEAT: DeviceError
DEVICE_ERROR_SYNTH_BAD_READ: DeviceError
DEVICE_ERROR_IF_DOWN_CONV_OVERHEAT: DeviceError
DEVICE_ERROR_IF_DOWN_CONV_BAD_READ: DeviceError
DEVICE_ERROR_RF_DOWN_CONV_OVERHEAT: DeviceError
DEVICE_ERROR_RF_DOWN_CONV_BAD_READ: DeviceError
DEVICE_ERROR_RF_UP_CONV_OVERHEAT: DeviceError
DEVICE_ERROR_RF_UP_CONV_BAD_READ: DeviceError
OCTAVE_MODULE_UNSPECIFIED: OctaveModule
OCTAVE_MODULE_RF_UPCONVERTER: OctaveModule
OCTAVE_MODULE_RF_DOWNCONVERTER: OctaveModule
OCTAVE_MODULE_IF_DOWNCONVERTER: OctaveModule
OCTAVE_MODULE_SYNTHESIZER: OctaveModule
OCTAVE_MODULE_MOTHERBOARD: OctaveModule
OCTAVE_MODULE_SOM: OctaveModule
CONSTANT_SOURCE_UNSPECIFIED: ConstantSource
CONSTANT_SOURCE_50_OHM: ConstantSource
CONSTANT_SOURCE_SHORT: ConstantSource
CONSTANT_SOURCE_OPEN: ConstantSource

class SubscribeStateRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SubscribeStateResponse(_message.Message):
    __slots__ = ("state", "error")
    STATE_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    state: DeviceState
    error: DeviceError
    def __init__(self, state: _Optional[_Union[DeviceState, str]] = ..., error: _Optional[_Union[DeviceError, str]] = ...) -> None: ...

class ResetRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ResetResponse(_message.Message):
    __slots__ = ("is_success", "error_message")
    IS_SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    is_success: bool
    error_message: str
    def __init__(self, is_success: bool = ..., error_message: _Optional[str] = ...) -> None: ...

class HealthRequest(_message.Message):
    __slots__ = ("monitor_interval_seconds", "stop_stream")
    MONITOR_INTERVAL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    STOP_STREAM_FIELD_NUMBER: _ClassVar[int]
    monitor_interval_seconds: int
    stop_stream: bool
    def __init__(self, monitor_interval_seconds: _Optional[int] = ..., stop_stream: bool = ...) -> None: ...

class HealthResponse(_message.Message):
    __slots__ = ("explore", "monitor")
    EXPLORE_FIELD_NUMBER: _ClassVar[int]
    MONITOR_FIELD_NUMBER: _ClassVar[int]
    explore: ExploreResponse
    monitor: MonitorResponse
    def __init__(self, explore: _Optional[_Union[ExploreResponse, _Mapping]] = ..., monitor: _Optional[_Union[MonitorResponse, _Mapping]] = ...) -> None: ...

class GetVersionRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetVersionResponse(_message.Message):
    __slots__ = ("version",)
    VERSION_FIELD_NUMBER: _ClassVar[int]
    version: str
    def __init__(self, version: _Optional[str] = ...) -> None: ...

class ModuleReference(_message.Message):
    __slots__ = ("type", "index")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    INDEX_FIELD_NUMBER: _ClassVar[int]
    type: OctaveModule
    index: int
    def __init__(self, type: _Optional[_Union[OctaveModule, str]] = ..., index: _Optional[int] = ...) -> None: ...

class SaveRequest(_message.Message):
    __slots__ = ("id", "modules", "overwrite", "timestamp")
    ID_FIELD_NUMBER: _ClassVar[int]
    MODULES_FIELD_NUMBER: _ClassVar[int]
    OVERWRITE_FIELD_NUMBER: _ClassVar[int]
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    id: str
    modules: _containers.RepeatedCompositeFieldContainer[ModuleReference]
    overwrite: bool
    timestamp: int
    def __init__(self, id: _Optional[str] = ..., modules: _Optional[_Iterable[_Union[ModuleReference, _Mapping]]] = ..., overwrite: bool = ..., timestamp: _Optional[int] = ...) -> None: ...

class SaveResponse(_message.Message):
    __slots__ = ("success", "message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    def __init__(self, success: bool = ..., message: _Optional[str] = ...) -> None: ...

class RecallRequest(_message.Message):
    __slots__ = ("id",)
    ID_FIELD_NUMBER: _ClassVar[int]
    id: str
    def __init__(self, id: _Optional[str] = ...) -> None: ...

class RecallResponse(_message.Message):
    __slots__ = ("success", "error_message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error_message: str
    def __init__(self, success: bool = ..., error_message: _Optional[str] = ...) -> None: ...

class ListRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class SaveInfo(_message.Message):
    __slots__ = ("request", "content")
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    request: SaveRequest
    content: UpdateRequest
    def __init__(self, request: _Optional[_Union[SaveRequest, _Mapping]] = ..., content: _Optional[_Union[UpdateRequest, _Mapping]] = ...) -> None: ...

class ListResponse(_message.Message):
    __slots__ = ("save_infos",)
    SAVE_INFOS_FIELD_NUMBER: _ClassVar[int]
    save_infos: _containers.RepeatedCompositeFieldContainer[SaveInfo]
    def __init__(self, save_infos: _Optional[_Iterable[_Union[SaveInfo, _Mapping]]] = ...) -> None: ...

class MonitorRequest(_message.Message):
    __slots__ = ("sense_only",)
    SENSE_ONLY_FIELD_NUMBER: _ClassVar[int]
    sense_only: bool
    def __init__(self, sense_only: bool = ...) -> None: ...

class MonitorResponse(_message.Message):
    __slots__ = ("modules",)
    class OctaveError(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        OCTAVE_ERROR_UNSPECIFIED: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_BAD_EXTERNAL_CLOCK: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_BAD_INTERNAL_CLOCK: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_FPGA_NOT_FOUND: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_FPGA_HW_FAIL: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_FPGA_BITFILE_FAIL: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_MODULE_NOT_FOUND: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_OVERHEAT: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_OVERHEAT_PROTECTION: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_READ_TEMP_FAIL: _ClassVar[MonitorResponse.OctaveError]
        OCTAVE_ERROR_PLL_NOT_LOCKED: _ClassVar[MonitorResponse.OctaveError]
    OCTAVE_ERROR_UNSPECIFIED: MonitorResponse.OctaveError
    OCTAVE_ERROR_BAD_EXTERNAL_CLOCK: MonitorResponse.OctaveError
    OCTAVE_ERROR_BAD_INTERNAL_CLOCK: MonitorResponse.OctaveError
    OCTAVE_ERROR_FPGA_NOT_FOUND: MonitorResponse.OctaveError
    OCTAVE_ERROR_FPGA_HW_FAIL: MonitorResponse.OctaveError
    OCTAVE_ERROR_FPGA_BITFILE_FAIL: MonitorResponse.OctaveError
    OCTAVE_ERROR_MODULE_NOT_FOUND: MonitorResponse.OctaveError
    OCTAVE_ERROR_OVERHEAT: MonitorResponse.OctaveError
    OCTAVE_ERROR_OVERHEAT_PROTECTION: MonitorResponse.OctaveError
    OCTAVE_ERROR_READ_TEMP_FAIL: MonitorResponse.OctaveError
    OCTAVE_ERROR_PLL_NOT_LOCKED: MonitorResponse.OctaveError
    class ModuleStatusError(_message.Message):
        __slots__ = ("type",)
        TYPE_FIELD_NUMBER: _ClassVar[int]
        type: MonitorResponse.OctaveError
        def __init__(self, type: _Optional[_Union[MonitorResponse.OctaveError, str]] = ...) -> None: ...
    class ModuleStatus(_message.Message):
        __slots__ = ("module", "temperature", "errors")
        MODULE_FIELD_NUMBER: _ClassVar[int]
        TEMPERATURE_FIELD_NUMBER: _ClassVar[int]
        ERRORS_FIELD_NUMBER: _ClassVar[int]
        module: ModuleReference
        temperature: float
        errors: _containers.RepeatedCompositeFieldContainer[MonitorResponse.ModuleStatusError]
        def __init__(self, module: _Optional[_Union[ModuleReference, _Mapping]] = ..., temperature: _Optional[float] = ..., errors: _Optional[_Iterable[_Union[MonitorResponse.ModuleStatusError, _Mapping]]] = ...) -> None: ...
    MODULES_FIELD_NUMBER: _ClassVar[int]
    modules: _containers.RepeatedCompositeFieldContainer[MonitorResponse.ModuleStatus]
    def __init__(self, modules: _Optional[_Iterable[_Union[MonitorResponse.ModuleStatus, _Mapping]]] = ...) -> None: ...

class ControlRequest(_message.Message):
    __slots__ = ("w_data", "r_length")
    W_DATA_FIELD_NUMBER: _ClassVar[int]
    R_LENGTH_FIELD_NUMBER: _ClassVar[int]
    w_data: bytes
    r_length: int
    def __init__(self, w_data: _Optional[bytes] = ..., r_length: _Optional[int] = ...) -> None: ...

class ControlResponse(_message.Message):
    __slots__ = ("r_data",)
    class RdataDebug(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        RDATA_DEBUG_UNSPECIFIED: _ClassVar[ControlResponse.RdataDebug]
        RDATA_DEBUG_ERROR_RESPONSE: _ClassVar[ControlResponse.RdataDebug]
        RDATA_DEBUG_SUCCESS_RESPONSE: _ClassVar[ControlResponse.RdataDebug]
    RDATA_DEBUG_UNSPECIFIED: ControlResponse.RdataDebug
    RDATA_DEBUG_ERROR_RESPONSE: ControlResponse.RdataDebug
    RDATA_DEBUG_SUCCESS_RESPONSE: ControlResponse.RdataDebug
    R_DATA_FIELD_NUMBER: _ClassVar[int]
    r_data: bytes
    def __init__(self, r_data: _Optional[bytes] = ...) -> None: ...

class ExploreRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ExploreResponse(_message.Message):
    __slots__ = ("modules",)
    class ModuleId(_message.Message):
        __slots__ = ("module", "id")
        MODULE_FIELD_NUMBER: _ClassVar[int]
        ID_FIELD_NUMBER: _ClassVar[int]
        module: ModuleReference
        id: str
        def __init__(self, module: _Optional[_Union[ModuleReference, _Mapping]] = ..., id: _Optional[str] = ...) -> None: ...
    MODULES_FIELD_NUMBER: _ClassVar[int]
    modules: _containers.RepeatedCompositeFieldContainer[ExploreResponse.ModuleId]
    def __init__(self, modules: _Optional[_Iterable[_Union[ExploreResponse.ModuleId, _Mapping]]] = ...) -> None: ...

class UpdateRequest(_message.Message):
    __slots__ = ("updates",)
    UPDATES_FIELD_NUMBER: _ClassVar[int]
    updates: _containers.RepeatedCompositeFieldContainer[SingleUpdate]
    def __init__(self, updates: _Optional[_Iterable[_Union[SingleUpdate, _Mapping]]] = ...) -> None: ...

class UpdateResponse(_message.Message):
    __slots__ = ("success", "error_message")
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    ERROR_MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    error_message: str
    def __init__(self, success: bool = ..., error_message: _Optional[str] = ...) -> None: ...

class SingleUpdate(_message.Message):
    __slots__ = ("rf_up_conv", "rf_down_conv", "if_down_conv", "synth", "clock", "motherboard")
    RF_UP_CONV_FIELD_NUMBER: _ClassVar[int]
    RF_DOWN_CONV_FIELD_NUMBER: _ClassVar[int]
    IF_DOWN_CONV_FIELD_NUMBER: _ClassVar[int]
    SYNTH_FIELD_NUMBER: _ClassVar[int]
    CLOCK_FIELD_NUMBER: _ClassVar[int]
    MOTHERBOARD_FIELD_NUMBER: _ClassVar[int]
    rf_up_conv: RFUpConvUpdate
    rf_down_conv: RFDownConvUpdate
    if_down_conv: IFDownConvUpdate
    synth: SynthUpdate
    clock: ClockUpdate
    motherboard: MotherboardUpdate
    def __init__(self, rf_up_conv: _Optional[_Union[RFUpConvUpdate, _Mapping]] = ..., rf_down_conv: _Optional[_Union[RFDownConvUpdate, _Mapping]] = ..., if_down_conv: _Optional[_Union[IFDownConvUpdate, _Mapping]] = ..., synth: _Optional[_Union[SynthUpdate, _Mapping]] = ..., clock: _Optional[_Union[ClockUpdate, _Mapping]] = ..., motherboard: _Optional[_Union[MotherboardUpdate, _Mapping]] = ...) -> None: ...

class RFUpConvUpdate(_message.Message):
    __slots__ = ("index", "input_attn", "enabled", "mixer_output_attn", "power_amp_enabled", "power_amp_attn", "fast_switch_mode")
    class IfInputSelection(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        IF_INPUT_SELECTION_UNSPECIFIED: _ClassVar[RFUpConvUpdate.IfInputSelection]
        IF_INPUT_SELECTION_PASS_THROUGH: _ClassVar[RFUpConvUpdate.IfInputSelection]
        IF_INPUT_SELECTION_ATTENUATE_10DB: _ClassVar[RFUpConvUpdate.IfInputSelection]
    IF_INPUT_SELECTION_UNSPECIFIED: RFUpConvUpdate.IfInputSelection
    IF_INPUT_SELECTION_PASS_THROUGH: RFUpConvUpdate.IfInputSelection
    IF_INPUT_SELECTION_ATTENUATE_10DB: RFUpConvUpdate.IfInputSelection
    class FastSwitchMode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        FAST_SWITCH_MODE_UNSPECIFIED: _ClassVar[RFUpConvUpdate.FastSwitchMode]
        FAST_SWITCH_MODE_ON: _ClassVar[RFUpConvUpdate.FastSwitchMode]
        FAST_SWITCH_MODE_OFF: _ClassVar[RFUpConvUpdate.FastSwitchMode]
        FAST_SWITCH_MODE_DIRECT: _ClassVar[RFUpConvUpdate.FastSwitchMode]
        FAST_SWITCH_MODE_INVERTED: _ClassVar[RFUpConvUpdate.FastSwitchMode]
    FAST_SWITCH_MODE_UNSPECIFIED: RFUpConvUpdate.FastSwitchMode
    FAST_SWITCH_MODE_ON: RFUpConvUpdate.FastSwitchMode
    FAST_SWITCH_MODE_OFF: RFUpConvUpdate.FastSwitchMode
    FAST_SWITCH_MODE_DIRECT: RFUpConvUpdate.FastSwitchMode
    FAST_SWITCH_MODE_INVERTED: RFUpConvUpdate.FastSwitchMode
    INDEX_FIELD_NUMBER: _ClassVar[int]
    INPUT_ATTN_FIELD_NUMBER: _ClassVar[int]
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    MIXER_OUTPUT_ATTN_FIELD_NUMBER: _ClassVar[int]
    POWER_AMP_ENABLED_FIELD_NUMBER: _ClassVar[int]
    POWER_AMP_ATTN_FIELD_NUMBER: _ClassVar[int]
    FAST_SWITCH_MODE_FIELD_NUMBER: _ClassVar[int]
    index: int
    input_attn: RFUpConvUpdate.IfInputSelection
    enabled: _wrappers_pb2.BoolValue
    mixer_output_attn: _wrappers_pb2.UInt32Value
    power_amp_enabled: _wrappers_pb2.BoolValue
    power_amp_attn: _wrappers_pb2.UInt32Value
    fast_switch_mode: RFUpConvUpdate.FastSwitchMode
    def __init__(self, index: _Optional[int] = ..., input_attn: _Optional[_Union[RFUpConvUpdate.IfInputSelection, str]] = ..., enabled: _Optional[_Union[_wrappers_pb2.BoolValue, _Mapping]] = ..., mixer_output_attn: _Optional[_Union[_wrappers_pb2.UInt32Value, _Mapping]] = ..., power_amp_enabled: _Optional[_Union[_wrappers_pb2.BoolValue, _Mapping]] = ..., power_amp_attn: _Optional[_Union[_wrappers_pb2.UInt32Value, _Mapping]] = ..., fast_switch_mode: _Optional[_Union[RFUpConvUpdate.FastSwitchMode, str]] = ...) -> None: ...

class RFDownConvUpdate(_message.Message):
    __slots__ = ("index", "rf_input", "lo_input", "enabled")
    class RFInput(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        RF_INPUT_UNSPECIFIED: _ClassVar[RFDownConvUpdate.RFInput]
        RF_INPUT_DEBUG_1: _ClassVar[RFDownConvUpdate.RFInput]
        RF_INPUT_DEBUG_2: _ClassVar[RFDownConvUpdate.RFInput]
        RF_INPUT_DEBUG_3: _ClassVar[RFDownConvUpdate.RFInput]
        RF_INPUT_DEBUG_4: _ClassVar[RFDownConvUpdate.RFInput]
        RF_INPUT_DEBUG_5: _ClassVar[RFDownConvUpdate.RFInput]
        RF_INPUT_MAIN: _ClassVar[RFDownConvUpdate.RFInput]
        RF_INPUT_FPGA_CONTROL: _ClassVar[RFDownConvUpdate.RFInput]
        RF_INPUT_DISCONNECT: _ClassVar[RFDownConvUpdate.RFInput]
    RF_INPUT_UNSPECIFIED: RFDownConvUpdate.RFInput
    RF_INPUT_DEBUG_1: RFDownConvUpdate.RFInput
    RF_INPUT_DEBUG_2: RFDownConvUpdate.RFInput
    RF_INPUT_DEBUG_3: RFDownConvUpdate.RFInput
    RF_INPUT_DEBUG_4: RFDownConvUpdate.RFInput
    RF_INPUT_DEBUG_5: RFDownConvUpdate.RFInput
    RF_INPUT_MAIN: RFDownConvUpdate.RFInput
    RF_INPUT_FPGA_CONTROL: RFDownConvUpdate.RFInput
    RF_INPUT_DISCONNECT: RFDownConvUpdate.RFInput
    class LOInput(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        LO_INPUT_UNSPECIFIED: _ClassVar[RFDownConvUpdate.LOInput]
        LO_INPUT_1: _ClassVar[RFDownConvUpdate.LOInput]
        LO_INPUT_2: _ClassVar[RFDownConvUpdate.LOInput]
        LO_INPUT_FPGA_CONTROL: _ClassVar[RFDownConvUpdate.LOInput]
    LO_INPUT_UNSPECIFIED: RFDownConvUpdate.LOInput
    LO_INPUT_1: RFDownConvUpdate.LOInput
    LO_INPUT_2: RFDownConvUpdate.LOInput
    LO_INPUT_FPGA_CONTROL: RFDownConvUpdate.LOInput
    INDEX_FIELD_NUMBER: _ClassVar[int]
    RF_INPUT_FIELD_NUMBER: _ClassVar[int]
    LO_INPUT_FIELD_NUMBER: _ClassVar[int]
    ENABLED_FIELD_NUMBER: _ClassVar[int]
    index: int
    rf_input: RFDownConvUpdate.RFInput
    lo_input: RFDownConvUpdate.LOInput
    enabled: _wrappers_pb2.BoolValue
    def __init__(self, index: _Optional[int] = ..., rf_input: _Optional[_Union[RFDownConvUpdate.RFInput, str]] = ..., lo_input: _Optional[_Union[RFDownConvUpdate.LOInput, str]] = ..., enabled: _Optional[_Union[_wrappers_pb2.BoolValue, _Mapping]] = ...) -> None: ...

class IFDownConvUpdate(_message.Message):
    __slots__ = ("index", "channel1", "channel2")
    class Mode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        MODE_UNSPECIFIED: _ClassVar[IFDownConvUpdate.Mode]
        MODE_OFF: _ClassVar[IFDownConvUpdate.Mode]
        MODE_BYPASS: _ClassVar[IFDownConvUpdate.Mode]
        MODE_POWER_DETECT: _ClassVar[IFDownConvUpdate.Mode]
        MODE_MIXER: _ClassVar[IFDownConvUpdate.Mode]
    MODE_UNSPECIFIED: IFDownConvUpdate.Mode
    MODE_OFF: IFDownConvUpdate.Mode
    MODE_BYPASS: IFDownConvUpdate.Mode
    MODE_POWER_DETECT: IFDownConvUpdate.Mode
    MODE_MIXER: IFDownConvUpdate.Mode
    class Coupling(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        COUPLING_UNSPECIFIED: _ClassVar[IFDownConvUpdate.Coupling]
        COUPLING_OPEN: _ClassVar[IFDownConvUpdate.Coupling]
        COUPLING_DC: _ClassVar[IFDownConvUpdate.Coupling]
        COUPLING_AC: _ClassVar[IFDownConvUpdate.Coupling]
    COUPLING_UNSPECIFIED: IFDownConvUpdate.Coupling
    COUPLING_OPEN: IFDownConvUpdate.Coupling
    COUPLING_DC: IFDownConvUpdate.Coupling
    COUPLING_AC: IFDownConvUpdate.Coupling
    class Channel(_message.Message):
        __slots__ = ("mode", "coupling")
        MODE_FIELD_NUMBER: _ClassVar[int]
        COUPLING_FIELD_NUMBER: _ClassVar[int]
        mode: IFDownConvUpdate.Mode
        coupling: IFDownConvUpdate.Coupling
        def __init__(self, mode: _Optional[_Union[IFDownConvUpdate.Mode, str]] = ..., coupling: _Optional[_Union[IFDownConvUpdate.Coupling, str]] = ...) -> None: ...
    INDEX_FIELD_NUMBER: _ClassVar[int]
    CHANNEL1_FIELD_NUMBER: _ClassVar[int]
    CHANNEL2_FIELD_NUMBER: _ClassVar[int]
    index: int
    channel1: IFDownConvUpdate.Channel
    channel2: IFDownConvUpdate.Channel
    def __init__(self, index: _Optional[int] = ..., channel1: _Optional[_Union[IFDownConvUpdate.Channel, _Mapping]] = ..., channel2: _Optional[_Union[IFDownConvUpdate.Channel, _Mapping]] = ...) -> None: ...

class ClockUpdate(_message.Message):
    __slots__ = ("mode", "clock_frequency", "synthesizers_clock", "clustered")
    class Mode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        MODE_UNSPECIFIED: _ClassVar[ClockUpdate.Mode]
        MODE_BUFFERED: _ClassVar[ClockUpdate.Mode]
        MODE_EXTERNAL: _ClassVar[ClockUpdate.Mode]
        MODE_INTERNAL: _ClassVar[ClockUpdate.Mode]
    MODE_UNSPECIFIED: ClockUpdate.Mode
    MODE_BUFFERED: ClockUpdate.Mode
    MODE_EXTERNAL: ClockUpdate.Mode
    MODE_INTERNAL: ClockUpdate.Mode
    MODE_FIELD_NUMBER: _ClassVar[int]
    CLOCK_FREQUENCY_FIELD_NUMBER: _ClassVar[int]
    SYNTHESIZERS_CLOCK_FIELD_NUMBER: _ClassVar[int]
    CLUSTERED_FIELD_NUMBER: _ClassVar[int]
    mode: ClockUpdate.Mode
    clock_frequency: _wrappers_pb2.DoubleValue
    synthesizers_clock: _wrappers_pb2.DoubleValue
    clustered: bool
    def __init__(self, mode: _Optional[_Union[ClockUpdate.Mode, str]] = ..., clock_frequency: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., synthesizers_clock: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., clustered: bool = ...) -> None: ...

class MotherboardUpdate(_message.Message):
    __slots__ = ("fan_speed",)
    FAN_SPEED_FIELD_NUMBER: _ClassVar[int]
    fan_speed: _wrappers_pb2.DoubleValue
    def __init__(self, fan_speed: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ...) -> None: ...

class SynthUpdate(_message.Message):
    __slots__ = ("index", "reference_clock", "synth_output", "synth_output_power", "heater", "gain", "digital_attn", "main_source", "main_output", "secondary_output", "stabilizer")
    class ReferenceSource(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        REFERENCE_SOURCE_UNSPECIFIED: _ClassVar[SynthUpdate.ReferenceSource]
        REFERENCE_SOURCE_INTERNAL: _ClassVar[SynthUpdate.ReferenceSource]
        REFERENCE_SOURCE_EXTERNAL: _ClassVar[SynthUpdate.ReferenceSource]
    REFERENCE_SOURCE_UNSPECIFIED: SynthUpdate.ReferenceSource
    REFERENCE_SOURCE_INTERNAL: SynthUpdate.ReferenceSource
    REFERENCE_SOURCE_EXTERNAL: SynthUpdate.ReferenceSource
    class SynthOutputPower(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        SYNTH_OUTPUT_POWER_UNSPECIFIED: _ClassVar[SynthUpdate.SynthOutputPower]
        SYNTH_OUTPUT_POWER_NEG4DB: _ClassVar[SynthUpdate.SynthOutputPower]
        SYNTH_OUTPUT_POWER_NEG1DB: _ClassVar[SynthUpdate.SynthOutputPower]
        SYNTH_OUTPUT_POWER_POS2DB: _ClassVar[SynthUpdate.SynthOutputPower]
        SYNTH_OUTPUT_POWER_POS5DB: _ClassVar[SynthUpdate.SynthOutputPower]
    SYNTH_OUTPUT_POWER_UNSPECIFIED: SynthUpdate.SynthOutputPower
    SYNTH_OUTPUT_POWER_NEG4DB: SynthUpdate.SynthOutputPower
    SYNTH_OUTPUT_POWER_NEG1DB: SynthUpdate.SynthOutputPower
    SYNTH_OUTPUT_POWER_POS2DB: SynthUpdate.SynthOutputPower
    SYNTH_OUTPUT_POWER_POS5DB: SynthUpdate.SynthOutputPower
    class MainSource(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        MAIN_SOURCE_UNSPECIFIED: _ClassVar[SynthUpdate.MainSource]
        MAIN_SOURCE_EXTERNAL: _ClassVar[SynthUpdate.MainSource]
        MAIN_SOURCE_SYNTHESIZER: _ClassVar[SynthUpdate.MainSource]
    MAIN_SOURCE_UNSPECIFIED: SynthUpdate.MainSource
    MAIN_SOURCE_EXTERNAL: SynthUpdate.MainSource
    MAIN_SOURCE_SYNTHESIZER: SynthUpdate.MainSource
    class MainOutput(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        MAIN_OUTPUT_UNSPECIFIED: _ClassVar[SynthUpdate.MainOutput]
        MAIN_OUTPUT_MAIN: _ClassVar[SynthUpdate.MainOutput]
        MAIN_OUTPUT_OFF: _ClassVar[SynthUpdate.MainOutput]
        MAIN_OUTPUT_FPGA_CONTROL: _ClassVar[SynthUpdate.MainOutput]
    MAIN_OUTPUT_UNSPECIFIED: SynthUpdate.MainOutput
    MAIN_OUTPUT_MAIN: SynthUpdate.MainOutput
    MAIN_OUTPUT_OFF: SynthUpdate.MainOutput
    MAIN_OUTPUT_FPGA_CONTROL: SynthUpdate.MainOutput
    class SecondaryOutput(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        SECONDARY_OUTPUT_UNSPECIFIED: _ClassVar[SynthUpdate.SecondaryOutput]
        SECONDARY_OUTPUT_MAIN: _ClassVar[SynthUpdate.SecondaryOutput]
        SECONDARY_OUTPUT_AUXILARY: _ClassVar[SynthUpdate.SecondaryOutput]
        SECONDARY_OUTPUT_OFF: _ClassVar[SynthUpdate.SecondaryOutput]
        SECONDARY_OUTPUT_FPGA_CONTROL: _ClassVar[SynthUpdate.SecondaryOutput]
    SECONDARY_OUTPUT_UNSPECIFIED: SynthUpdate.SecondaryOutput
    SECONDARY_OUTPUT_MAIN: SynthUpdate.SecondaryOutput
    SECONDARY_OUTPUT_AUXILARY: SynthUpdate.SecondaryOutput
    SECONDARY_OUTPUT_OFF: SynthUpdate.SecondaryOutput
    SECONDARY_OUTPUT_FPGA_CONTROL: SynthUpdate.SecondaryOutput
    class ReferenceClock(_message.Message):
        __slots__ = ("source", "frequency", "divider")
        SOURCE_FIELD_NUMBER: _ClassVar[int]
        FREQUENCY_FIELD_NUMBER: _ClassVar[int]
        DIVIDER_FIELD_NUMBER: _ClassVar[int]
        source: SynthUpdate.ReferenceSource
        frequency: _wrappers_pb2.DoubleValue
        divider: _wrappers_pb2.Int32Value
        def __init__(self, source: _Optional[_Union[SynthUpdate.ReferenceSource, str]] = ..., frequency: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., divider: _Optional[_Union[_wrappers_pb2.Int32Value, _Mapping]] = ...) -> None: ...
    class SynthOutput(_message.Message):
        __slots__ = ("frequency", "disabled")
        FREQUENCY_FIELD_NUMBER: _ClassVar[int]
        DISABLED_FIELD_NUMBER: _ClassVar[int]
        frequency: _wrappers_pb2.DoubleValue
        disabled: _wrappers_pb2.BoolValue
        def __init__(self, frequency: _Optional[_Union[_wrappers_pb2.DoubleValue, _Mapping]] = ..., disabled: _Optional[_Union[_wrappers_pb2.BoolValue, _Mapping]] = ...) -> None: ...
    class Stabilizer(_message.Message):
        __slots__ = ("k_p", "k_i", "k_d", "a", "b", "heater_max", "set_point", "enabled")
        K_P_FIELD_NUMBER: _ClassVar[int]
        K_I_FIELD_NUMBER: _ClassVar[int]
        K_D_FIELD_NUMBER: _ClassVar[int]
        A_FIELD_NUMBER: _ClassVar[int]
        B_FIELD_NUMBER: _ClassVar[int]
        HEATER_MAX_FIELD_NUMBER: _ClassVar[int]
        SET_POINT_FIELD_NUMBER: _ClassVar[int]
        ENABLED_FIELD_NUMBER: _ClassVar[int]
        k_p: _wrappers_pb2.FloatValue
        k_i: _wrappers_pb2.FloatValue
        k_d: _wrappers_pb2.FloatValue
        a: _wrappers_pb2.FloatValue
        b: _wrappers_pb2.FloatValue
        heater_max: _wrappers_pb2.UInt32Value
        set_point: _wrappers_pb2.FloatValue
        enabled: _wrappers_pb2.BoolValue
        def __init__(self, k_p: _Optional[_Union[_wrappers_pb2.FloatValue, _Mapping]] = ..., k_i: _Optional[_Union[_wrappers_pb2.FloatValue, _Mapping]] = ..., k_d: _Optional[_Union[_wrappers_pb2.FloatValue, _Mapping]] = ..., a: _Optional[_Union[_wrappers_pb2.FloatValue, _Mapping]] = ..., b: _Optional[_Union[_wrappers_pb2.FloatValue, _Mapping]] = ..., heater_max: _Optional[_Union[_wrappers_pb2.UInt32Value, _Mapping]] = ..., set_point: _Optional[_Union[_wrappers_pb2.FloatValue, _Mapping]] = ..., enabled: _Optional[_Union[_wrappers_pb2.BoolValue, _Mapping]] = ...) -> None: ...
    INDEX_FIELD_NUMBER: _ClassVar[int]
    REFERENCE_CLOCK_FIELD_NUMBER: _ClassVar[int]
    SYNTH_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    SYNTH_OUTPUT_POWER_FIELD_NUMBER: _ClassVar[int]
    HEATER_FIELD_NUMBER: _ClassVar[int]
    GAIN_FIELD_NUMBER: _ClassVar[int]
    DIGITAL_ATTN_FIELD_NUMBER: _ClassVar[int]
    MAIN_SOURCE_FIELD_NUMBER: _ClassVar[int]
    MAIN_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    SECONDARY_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    STABILIZER_FIELD_NUMBER: _ClassVar[int]
    index: int
    reference_clock: SynthUpdate.ReferenceClock
    synth_output: SynthUpdate.SynthOutput
    synth_output_power: SynthUpdate.SynthOutputPower
    heater: _wrappers_pb2.UInt32Value
    gain: _wrappers_pb2.UInt32Value
    digital_attn: _wrappers_pb2.UInt32Value
    main_source: SynthUpdate.MainSource
    main_output: SynthUpdate.MainOutput
    secondary_output: SynthUpdate.SecondaryOutput
    stabilizer: SynthUpdate.Stabilizer
    def __init__(self, index: _Optional[int] = ..., reference_clock: _Optional[_Union[SynthUpdate.ReferenceClock, _Mapping]] = ..., synth_output: _Optional[_Union[SynthUpdate.SynthOutput, _Mapping]] = ..., synth_output_power: _Optional[_Union[SynthUpdate.SynthOutputPower, str]] = ..., heater: _Optional[_Union[_wrappers_pb2.UInt32Value, _Mapping]] = ..., gain: _Optional[_Union[_wrappers_pb2.UInt32Value, _Mapping]] = ..., digital_attn: _Optional[_Union[_wrappers_pb2.UInt32Value, _Mapping]] = ..., main_source: _Optional[_Union[SynthUpdate.MainSource, str]] = ..., main_output: _Optional[_Union[SynthUpdate.MainOutput, str]] = ..., secondary_output: _Optional[_Union[SynthUpdate.SecondaryOutput, str]] = ..., stabilizer: _Optional[_Union[SynthUpdate.Stabilizer, _Mapping]] = ...) -> None: ...

class AquireRequest(_message.Message):
    __slots__ = ("modules", "use_cache")
    MODULES_FIELD_NUMBER: _ClassVar[int]
    USE_CACHE_FIELD_NUMBER: _ClassVar[int]
    modules: _containers.RepeatedCompositeFieldContainer[ModuleReference]
    use_cache: bool
    def __init__(self, modules: _Optional[_Iterable[_Union[ModuleReference, _Mapping]]] = ..., use_cache: bool = ...) -> None: ...

class AquireResponse(_message.Message):
    __slots__ = ("state",)
    STATE_FIELD_NUMBER: _ClassVar[int]
    state: UpdateRequest
    def __init__(self, state: _Optional[_Union[UpdateRequest, _Mapping]] = ...) -> None: ...

class IdentifyRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class IdentifyResponse(_message.Message):
    __slots__ = ("rf_up_converters", "rf_down_converters", "if_down_converters", "synthesizers", "panel_identity")
    RF_UP_CONVERTERS_FIELD_NUMBER: _ClassVar[int]
    RF_DOWN_CONVERTERS_FIELD_NUMBER: _ClassVar[int]
    IF_DOWN_CONVERTERS_FIELD_NUMBER: _ClassVar[int]
    SYNTHESIZERS_FIELD_NUMBER: _ClassVar[int]
    PANEL_IDENTITY_FIELD_NUMBER: _ClassVar[int]
    rf_up_converters: _containers.RepeatedCompositeFieldContainer[RFUpConvIdentity]
    rf_down_converters: _containers.RepeatedCompositeFieldContainer[RFDownConvIdentity]
    if_down_converters: _containers.RepeatedCompositeFieldContainer[IFDownConvIdentity]
    synthesizers: _containers.RepeatedCompositeFieldContainer[SynthIdentity]
    panel_identity: PanelIdentity
    def __init__(self, rf_up_converters: _Optional[_Iterable[_Union[RFUpConvIdentity, _Mapping]]] = ..., rf_down_converters: _Optional[_Iterable[_Union[RFDownConvIdentity, _Mapping]]] = ..., if_down_converters: _Optional[_Iterable[_Union[IFDownConvIdentity, _Mapping]]] = ..., synthesizers: _Optional[_Iterable[_Union[SynthIdentity, _Mapping]]] = ..., panel_identity: _Optional[_Union[PanelIdentity, _Mapping]] = ...) -> None: ...

class UpConvRFOutput(_message.Message):
    __slots__ = ("index",)
    INDEX_FIELD_NUMBER: _ClassVar[int]
    index: int
    def __init__(self, index: _Optional[int] = ...) -> None: ...

class SynthRFOutput(_message.Message):
    __slots__ = ("index", "output_port")
    class OutputPort(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        OUTPUT_PORT_UNSPECIFIED: _ClassVar[SynthRFOutput.OutputPort]
        OUTPUT_PORT_SYNTH: _ClassVar[SynthRFOutput.OutputPort]
        OUTPUT_PORT_MAIN: _ClassVar[SynthRFOutput.OutputPort]
        OUTPUT_PORT_SECONDARY: _ClassVar[SynthRFOutput.OutputPort]
    OUTPUT_PORT_UNSPECIFIED: SynthRFOutput.OutputPort
    OUTPUT_PORT_SYNTH: SynthRFOutput.OutputPort
    OUTPUT_PORT_MAIN: SynthRFOutput.OutputPort
    OUTPUT_PORT_SECONDARY: SynthRFOutput.OutputPort
    INDEX_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_PORT_FIELD_NUMBER: _ClassVar[int]
    index: int
    output_port: SynthRFOutput.OutputPort
    def __init__(self, index: _Optional[int] = ..., output_port: _Optional[_Union[SynthRFOutput.OutputPort, str]] = ...) -> None: ...

class ExternalRFInput(_message.Message):
    __slots__ = ("lo_input_index", "demod_lo_input_index", "rf_in_index")
    LO_INPUT_INDEX_FIELD_NUMBER: _ClassVar[int]
    DEMOD_LO_INPUT_INDEX_FIELD_NUMBER: _ClassVar[int]
    RF_IN_INDEX_FIELD_NUMBER: _ClassVar[int]
    lo_input_index: int
    demod_lo_input_index: int
    rf_in_index: int
    def __init__(self, lo_input_index: _Optional[int] = ..., demod_lo_input_index: _Optional[int] = ..., rf_in_index: _Optional[int] = ...) -> None: ...

class RFSource(_message.Message):
    __slots__ = ("rf_up_conv_output", "synth_output", "external_input", "constant_source")
    RF_UP_CONV_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    SYNTH_OUTPUT_FIELD_NUMBER: _ClassVar[int]
    EXTERNAL_INPUT_FIELD_NUMBER: _ClassVar[int]
    CONSTANT_SOURCE_FIELD_NUMBER: _ClassVar[int]
    rf_up_conv_output: UpConvRFOutput
    synth_output: SynthRFOutput
    external_input: ExternalRFInput
    constant_source: ConstantSource
    def __init__(self, rf_up_conv_output: _Optional[_Union[UpConvRFOutput, _Mapping]] = ..., synth_output: _Optional[_Union[SynthRFOutput, _Mapping]] = ..., external_input: _Optional[_Union[ExternalRFInput, _Mapping]] = ..., constant_source: _Optional[_Union[ConstantSource, str]] = ...) -> None: ...

class RFDownConvIFSource(_message.Message):
    __slots__ = ("index", "output_port")
    class OutputPort(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        OUTPUT_PORT_UNSPECIFIED: _ClassVar[RFDownConvIFSource.OutputPort]
        OUTPUT_PORT_I: _ClassVar[RFDownConvIFSource.OutputPort]
        OUTPUT_PORT_Q: _ClassVar[RFDownConvIFSource.OutputPort]
    OUTPUT_PORT_UNSPECIFIED: RFDownConvIFSource.OutputPort
    OUTPUT_PORT_I: RFDownConvIFSource.OutputPort
    OUTPUT_PORT_Q: RFDownConvIFSource.OutputPort
    INDEX_FIELD_NUMBER: _ClassVar[int]
    OUTPUT_PORT_FIELD_NUMBER: _ClassVar[int]
    index: int
    output_port: RFDownConvIFSource.OutputPort
    def __init__(self, index: _Optional[int] = ..., output_port: _Optional[_Union[RFDownConvIFSource.OutputPort, str]] = ...) -> None: ...

class IFDownConvIFSource(_message.Message):
    __slots__ = ("index", "channel_index")
    INDEX_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_INDEX_FIELD_NUMBER: _ClassVar[int]
    index: int
    channel_index: int
    def __init__(self, index: _Optional[int] = ..., channel_index: _Optional[int] = ...) -> None: ...

class ExternalIFInput(_message.Message):
    __slots__ = ("i_index", "q_index", "if_lo_i_index", "if_lo_q_index")
    I_INDEX_FIELD_NUMBER: _ClassVar[int]
    Q_INDEX_FIELD_NUMBER: _ClassVar[int]
    IF_LO_I_INDEX_FIELD_NUMBER: _ClassVar[int]
    IF_LO_Q_INDEX_FIELD_NUMBER: _ClassVar[int]
    i_index: int
    q_index: int
    if_lo_i_index: int
    if_lo_q_index: int
    def __init__(self, i_index: _Optional[int] = ..., q_index: _Optional[int] = ..., if_lo_i_index: _Optional[int] = ..., if_lo_q_index: _Optional[int] = ...) -> None: ...

class IFSource(_message.Message):
    __slots__ = ("rf_downconv_if_source", "if_downconv_if_source", "external_if_input", "constant_source")
    RF_DOWNCONV_IF_SOURCE_FIELD_NUMBER: _ClassVar[int]
    IF_DOWNCONV_IF_SOURCE_FIELD_NUMBER: _ClassVar[int]
    EXTERNAL_IF_INPUT_FIELD_NUMBER: _ClassVar[int]
    CONSTANT_SOURCE_FIELD_NUMBER: _ClassVar[int]
    rf_downconv_if_source: RFDownConvIFSource
    if_downconv_if_source: IFDownConvIFSource
    external_if_input: ExternalIFInput
    constant_source: ConstantSource
    def __init__(self, rf_downconv_if_source: _Optional[_Union[RFDownConvIFSource, _Mapping]] = ..., if_downconv_if_source: _Optional[_Union[IFDownConvIFSource, _Mapping]] = ..., external_if_input: _Optional[_Union[ExternalIFInput, _Mapping]] = ..., constant_source: _Optional[_Union[ConstantSource, str]] = ...) -> None: ...

class RFUpConvIdentity(_message.Message):
    __slots__ = ("index", "connectivity", "parameters")
    class Connectivity(_message.Message):
        __slots__ = ("i_input", "q_input", "lo_input")
        I_INPUT_FIELD_NUMBER: _ClassVar[int]
        Q_INPUT_FIELD_NUMBER: _ClassVar[int]
        LO_INPUT_FIELD_NUMBER: _ClassVar[int]
        i_input: IFSource
        q_input: IFSource
        lo_input: RFSource
        def __init__(self, i_input: _Optional[_Union[IFSource, _Mapping]] = ..., q_input: _Optional[_Union[IFSource, _Mapping]] = ..., lo_input: _Optional[_Union[RFSource, _Mapping]] = ...) -> None: ...
    class Parameters(_message.Message):
        __slots__ = ("attn_1_db", "attn_2_db")
        ATTN_1_DB_FIELD_NUMBER: _ClassVar[int]
        ATTN_2_DB_FIELD_NUMBER: _ClassVar[int]
        attn_1_db: float
        attn_2_db: float
        def __init__(self, attn_1_db: _Optional[float] = ..., attn_2_db: _Optional[float] = ...) -> None: ...
    INDEX_FIELD_NUMBER: _ClassVar[int]
    CONNECTIVITY_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    index: int
    connectivity: RFUpConvIdentity.Connectivity
    parameters: RFUpConvIdentity.Parameters
    def __init__(self, index: _Optional[int] = ..., connectivity: _Optional[_Union[RFUpConvIdentity.Connectivity, _Mapping]] = ..., parameters: _Optional[_Union[RFUpConvIdentity.Parameters, _Mapping]] = ...) -> None: ...

class RFDownConvIdentity(_message.Message):
    __slots__ = ("index", "connectivity")
    class Connectivity(_message.Message):
        __slots__ = ("debug_rf_input_1", "debug_rf_input_2", "debug_rf_input_3", "debug_rf_input_4", "debug_rf_input_5", "rf_main_input", "lo_input_1", "lo_input_2")
        DEBUG_RF_INPUT_1_FIELD_NUMBER: _ClassVar[int]
        DEBUG_RF_INPUT_2_FIELD_NUMBER: _ClassVar[int]
        DEBUG_RF_INPUT_3_FIELD_NUMBER: _ClassVar[int]
        DEBUG_RF_INPUT_4_FIELD_NUMBER: _ClassVar[int]
        DEBUG_RF_INPUT_5_FIELD_NUMBER: _ClassVar[int]
        RF_MAIN_INPUT_FIELD_NUMBER: _ClassVar[int]
        LO_INPUT_1_FIELD_NUMBER: _ClassVar[int]
        LO_INPUT_2_FIELD_NUMBER: _ClassVar[int]
        debug_rf_input_1: RFSource
        debug_rf_input_2: RFSource
        debug_rf_input_3: RFSource
        debug_rf_input_4: RFSource
        debug_rf_input_5: RFSource
        rf_main_input: RFSource
        lo_input_1: RFSource
        lo_input_2: RFSource
        def __init__(self, debug_rf_input_1: _Optional[_Union[RFSource, _Mapping]] = ..., debug_rf_input_2: _Optional[_Union[RFSource, _Mapping]] = ..., debug_rf_input_3: _Optional[_Union[RFSource, _Mapping]] = ..., debug_rf_input_4: _Optional[_Union[RFSource, _Mapping]] = ..., debug_rf_input_5: _Optional[_Union[RFSource, _Mapping]] = ..., rf_main_input: _Optional[_Union[RFSource, _Mapping]] = ..., lo_input_1: _Optional[_Union[RFSource, _Mapping]] = ..., lo_input_2: _Optional[_Union[RFSource, _Mapping]] = ...) -> None: ...
    INDEX_FIELD_NUMBER: _ClassVar[int]
    CONNECTIVITY_FIELD_NUMBER: _ClassVar[int]
    index: int
    connectivity: RFDownConvIdentity.Connectivity
    def __init__(self, index: _Optional[int] = ..., connectivity: _Optional[_Union[RFDownConvIdentity.Connectivity, _Mapping]] = ...) -> None: ...

class IFDownConvIdentity(_message.Message):
    __slots__ = ("index", "connectivity")
    class Connectivity(_message.Message):
        __slots__ = ("channel_1_input", "channel_1_lo_input", "channel_2_input", "channel_2_lo_input")
        CHANNEL_1_INPUT_FIELD_NUMBER: _ClassVar[int]
        CHANNEL_1_LO_INPUT_FIELD_NUMBER: _ClassVar[int]
        CHANNEL_2_INPUT_FIELD_NUMBER: _ClassVar[int]
        CHANNEL_2_LO_INPUT_FIELD_NUMBER: _ClassVar[int]
        channel_1_input: IFSource
        channel_1_lo_input: IFSource
        channel_2_input: IFSource
        channel_2_lo_input: IFSource
        def __init__(self, channel_1_input: _Optional[_Union[IFSource, _Mapping]] = ..., channel_1_lo_input: _Optional[_Union[IFSource, _Mapping]] = ..., channel_2_input: _Optional[_Union[IFSource, _Mapping]] = ..., channel_2_lo_input: _Optional[_Union[IFSource, _Mapping]] = ...) -> None: ...
    INDEX_FIELD_NUMBER: _ClassVar[int]
    CONNECTIVITY_FIELD_NUMBER: _ClassVar[int]
    index: int
    connectivity: IFDownConvIdentity.Connectivity
    def __init__(self, index: _Optional[int] = ..., connectivity: _Optional[_Union[IFDownConvIdentity.Connectivity, _Mapping]] = ...) -> None: ...

class SynthIdentity(_message.Message):
    __slots__ = ("index", "connectivity", "parameters")
    class Connectivity(_message.Message):
        __slots__ = ("main_lo_input", "secondary_lo_input")
        MAIN_LO_INPUT_FIELD_NUMBER: _ClassVar[int]
        SECONDARY_LO_INPUT_FIELD_NUMBER: _ClassVar[int]
        main_lo_input: RFSource
        secondary_lo_input: RFSource
        def __init__(self, main_lo_input: _Optional[_Union[RFSource, _Mapping]] = ..., secondary_lo_input: _Optional[_Union[RFSource, _Mapping]] = ...) -> None: ...
    class Parameters(_message.Message):
        __slots__ = ("low_frequency_filters", "medium_frequency_filter", "high_frequency_filter")
        class LowFrequencyFilter(_message.Message):
            __slots__ = ("index", "filter_1", "filter_2")
            INDEX_FIELD_NUMBER: _ClassVar[int]
            FILTER_1_FIELD_NUMBER: _ClassVar[int]
            FILTER_2_FIELD_NUMBER: _ClassVar[int]
            index: int
            filter_1: str
            filter_2: str
            def __init__(self, index: _Optional[int] = ..., filter_1: _Optional[str] = ..., filter_2: _Optional[str] = ...) -> None: ...
        class ParametrizedFilter(_message.Message):
            __slots__ = ()
            def __init__(self) -> None: ...
        LOW_FREQUENCY_FILTERS_FIELD_NUMBER: _ClassVar[int]
        MEDIUM_FREQUENCY_FILTER_FIELD_NUMBER: _ClassVar[int]
        HIGH_FREQUENCY_FILTER_FIELD_NUMBER: _ClassVar[int]
        low_frequency_filters: _containers.RepeatedCompositeFieldContainer[SynthIdentity.Parameters.LowFrequencyFilter]
        medium_frequency_filter: SynthIdentity.Parameters.ParametrizedFilter
        high_frequency_filter: SynthIdentity.Parameters.ParametrizedFilter
        def __init__(self, low_frequency_filters: _Optional[_Iterable[_Union[SynthIdentity.Parameters.LowFrequencyFilter, _Mapping]]] = ..., medium_frequency_filter: _Optional[_Union[SynthIdentity.Parameters.ParametrizedFilter, _Mapping]] = ..., high_frequency_filter: _Optional[_Union[SynthIdentity.Parameters.ParametrizedFilter, _Mapping]] = ...) -> None: ...
    INDEX_FIELD_NUMBER: _ClassVar[int]
    CONNECTIVITY_FIELD_NUMBER: _ClassVar[int]
    PARAMETERS_FIELD_NUMBER: _ClassVar[int]
    index: int
    connectivity: SynthIdentity.Connectivity
    parameters: SynthIdentity.Parameters
    def __init__(self, index: _Optional[int] = ..., connectivity: _Optional[_Union[SynthIdentity.Connectivity, _Mapping]] = ..., parameters: _Optional[_Union[SynthIdentity.Parameters, _Mapping]] = ...) -> None: ...

class PanelIdentity(_message.Message):
    __slots__ = ("rf_outputs", "if_output_i", "if_output_q", "synth_outputs")
    class RFOutput(_message.Message):
        __slots__ = ("index", "source")
        INDEX_FIELD_NUMBER: _ClassVar[int]
        SOURCE_FIELD_NUMBER: _ClassVar[int]
        index: int
        source: RFSource
        def __init__(self, index: _Optional[int] = ..., source: _Optional[_Union[RFSource, _Mapping]] = ...) -> None: ...
    class SynthOutput(_message.Message):
        __slots__ = ("index", "source")
        INDEX_FIELD_NUMBER: _ClassVar[int]
        SOURCE_FIELD_NUMBER: _ClassVar[int]
        index: int
        source: RFSource
        def __init__(self, index: _Optional[int] = ..., source: _Optional[_Union[RFSource, _Mapping]] = ...) -> None: ...
    RF_OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    IF_OUTPUT_I_FIELD_NUMBER: _ClassVar[int]
    IF_OUTPUT_Q_FIELD_NUMBER: _ClassVar[int]
    SYNTH_OUTPUTS_FIELD_NUMBER: _ClassVar[int]
    rf_outputs: _containers.RepeatedCompositeFieldContainer[PanelIdentity.RFOutput]
    if_output_i: _containers.RepeatedCompositeFieldContainer[IFSource]
    if_output_q: _containers.RepeatedCompositeFieldContainer[IFSource]
    synth_outputs: _containers.RepeatedCompositeFieldContainer[PanelIdentity.SynthOutput]
    def __init__(self, rf_outputs: _Optional[_Iterable[_Union[PanelIdentity.RFOutput, _Mapping]]] = ..., if_output_i: _Optional[_Iterable[_Union[IFSource, _Mapping]]] = ..., if_output_q: _Optional[_Iterable[_Union[IFSource, _Mapping]]] = ..., synth_outputs: _Optional[_Iterable[_Union[PanelIdentity.SynthOutput, _Mapping]]] = ...) -> None: ...

class MotherboardStatus(_message.Message):
    __slots__ = ("valid_1g_vco",)
    VALID_1G_VCO_FIELD_NUMBER: _ClassVar[int]
    valid_1g_vco: bool
    def __init__(self, valid_1g_vco: bool = ...) -> None: ...



from typing import Any
from collections.abc import Mapping, Callable

from google.protobuf.message import Message
from google._upb._message import RepeatedCompositeContainer

from qm.utils import list_fields
from qm.utils.protobuf_utils import Node


class QuaNodeVisitor:
    def _default_enter(self, node: Message) -> bool:
        return True

    def _default_leave(self, node: Message) -> None:
        return

    def _default_visit(self, node: Message) -> None:
        for field in list_fields(node).values():
            self.visit(field)

    @property
    def _node_to_enter(self) -> Mapping[type, Callable[[Any], bool]]:
        return {}

    @property
    def _node_to_visit(self) -> Mapping[type, Callable[[Any], None]]:
        return {}

    @property
    def _node_to_leave(self) -> Mapping[type, Callable[[Any], None]]:
        return {}

    def _call_enter(self, node: Message) -> bool:
        if type(node) in self._node_to_enter:
            func = self._node_to_enter[type(node)]
            return func(node)
        return self._default_enter(node)

    def _call_visit(self, node: Message) -> None:
        if type(node) in self._node_to_visit:
            func = self._node_to_visit[type(node)]
            func(node)
        else:
            self._default_visit(node)

    def _call_leave(self, node: Message) -> None:
        if type(node) in self._node_to_leave:
            func = self._node_to_leave[type(node)]
            func(node)
        else:
            self._default_leave(node)

    def visit(self, node: Node) -> None:
        if isinstance(node, Message):
            self._enter_visit_leave(node)
        elif isinstance(node, RepeatedCompositeContainer):
            for n in node:
                self.visit(n)

    def _enter_visit_leave(self, node: Message) -> None:
        should_visit = self._call_enter(node)
        if should_visit:
            self._call_visit(node)
        self._call_leave(node)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qm/pb/general_messages.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'qm/pb/general_messages.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1cqm/pb/general_messages.proto\x12\x18qm.grpc.general_messages\"-\n\x0c\x45rrorMessage\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\"<\n\x06Matrix\x12\x0b\n\x03v00\x18\x01 \x01(\x01\x12\x0b\n\x03v01\x18\x02 \x01(\x01\x12\x0b\n\x03v10\x18\x03 \x01(\x01\x12\x0b\n\x03v11\x18\x04 \x01(\x01*Z\n\x0cMessageLevel\x12\x17\n\x13Message_LEVEL_ERROR\x10\x00\x12\x19\n\x15Message_LEVEL_WARNING\x10\x01\x12\x16\n\x12Message_LEVEL_INFO\x10\x02\x42\x37\n\x17qm.grpc.generalMessagesB\x14GeneralMessagesProtoP\x01\xa2\x02\x03HLWb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qm.pb.general_messages_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\027qm.grpc.generalMessagesB\024GeneralMessagesProtoP\001\242\002\003HLW'
  _globals['_MESSAGELEVEL']._serialized_start=167
  _globals['_MESSAGELEVEL']._serialized_end=257
  _globals['_ERRORMESSAGE']._serialized_start=58
  _globals['_ERRORMESSAGE']._serialized_end=103
  _globals['_MATRIX']._serialized_start=105
  _globals['_MATRIX']._serialized_end=165
# @@protoc_insertion_point(module_scope)



from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class MessageLevel(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    Message_LEVEL_ERROR: _ClassVar[MessageLevel]
    Message_LEVEL_WARNING: _ClassVar[MessageLevel]
    Message_LEVEL_INFO: _ClassVar[MessageLevel]
Message_LEVEL_ERROR: MessageLevel
Message_LEVEL_WARNING: MessageLevel
Message_LEVEL_INFO: MessageLevel

class ErrorMessage(_message.Message):
    __slots__ = ("code", "message")
    CODE_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    code: int
    message: str
    def __init__(self, code: _Optional[int] = ..., message: _Optional[str] = ...) -> None: ...

class Matrix(_message.Message):
    __slots__ = ("v00", "v01", "v10", "v11")
    V00_FIELD_NUMBER: _ClassVar[int]
    V01_FIELD_NUMBER: _ClassVar[int]
    V10_FIELD_NUMBER: _ClassVar[int]
    V11_FIELD_NUMBER: _ClassVar[int]
    v00: float
    v01: float
    v10: float
    v11: float
    def __init__(self, v00: _Optional[float] = ..., v01: _Optional[float] = ..., v10: _Optional[float] = ..., v11: _Optional[float] = ...) -> None: ...



from typing import Tuple, BinaryIO

import numpy.typing
from numpy.lib import format as _format


def _write_header(writer: BinaryIO, shape: Tuple[int, ...], d_type: object) -> None:
    corrected_dtype = _fix_unsupported_dtype(d_type)
    _format.write_array_header_2_0(writer, {"descr": corrected_dtype, "fortran_order": False, "shape": shape})  # type: ignore[no-untyped-call]


_NP_BOOL = numpy.dtype(numpy.bool_).str
_UNSUPPORTED_DTYPE = "bool8"


def _fix_unsupported_dtype(d_type: object) -> object:
    # Numpy2 stopped support for bool8, so we need to convert it to a valid bool type
    if d_type == _UNSUPPORTED_DTYPE:
        return _NP_BOOL
    if isinstance(d_type, list):
        return [_fix_unsupported_dtype(elem) for elem in d_type]
    return d_type



from typing import Union

from qm._loc import _get_loc
from qm.grpc.qm.pb import inc_qua_pb2
from qm.api.models.capabilities import QopCaps
from qm.qua._expressions import Scalar, to_scalar_pb_expression
from qm.qua._scope_management.scopes_manager import scopes_manager


def frame_rotation(angle: Union[Scalar[float]], *elements: str) -> None:
    r"""Shift the phase of the oscillator associated with an element by the given angle.

    This is typically used for virtual z-rotations.

    Note:
        The fixed point format of QUA variables of type fixed is 4.28, meaning the phase
        must be between $-8$ and $8-2^{28}$. Otherwise the phase value will be invalid.
        It is therefore better to use `frame_rotation_2pi()` which avoids this issue.

    Note:
        The phase is accumulated with a resolution of 16 bit.
        Therefore, *N* changes to the phase can result in a phase (and amplitude) inaccuracy of about :math:`N \cdot 2^{-16}`.
        To null out this accumulated error, it is recommended to use `reset_frame(el)` from time to time.

    Args:
        angle (Union[float, QUA variable of type fixed]): The angle to
            add to the current phase (in radians)
        *elements (str): a single element whose oscillator's phase will
            be shifted. multiple elements can be given, in which case
            all of their oscillators' phases will be shifted

    """
    if isinstance(angle, (inc_qua_pb2.QuaProgram.ArrayVarRefExpression, inc_qua_pb2.QuaProgram.VarRefExpression)):
        raise TypeError(f"angle cannot be of type {type(angle)}")
    frame_rotation_2pi(angle * 0.15915494309189535, *elements)


def frame_rotation_2pi(angle: Scalar[float], *elements: str) -> None:
    r"""Shift the phase of the oscillator associated with an element by the given angle in units of 2pi radians.

    This is typically used for virtual z-rotations.

    Note:
        Unlike the case of frame_rotation(), this method performs the 2-pi radian wrap around of the angle automatically.

    Note:
        The phase is accumulated with a resolution of 16 bit.
        Therefore, *N* changes to the phase can result in a phase inaccuracy of about :math:`N \cdot 2^{-16}`.
        To null out this accumulated error, it is recommended to use `reset_frame(el)` from time to time.

    Args:
        angle (Union[float,QUA variable of type real]): The angle to add
            to the current phase (in $2\pi$ radians)
        *elements (str): a single element whose oscillator's phase will
            be shifted. multiple elements can be given, in which case
            all of their oscillators' phases will be shifted

    """
    loc = _get_loc()
    for element in elements:
        statement = inc_qua_pb2.QuaProgram.AnyStatement(
            zRotation=inc_qua_pb2.QuaProgram.ZRotationStatement(
                loc=loc,
                value=to_scalar_pb_expression(angle),
                qe=inc_qua_pb2.QuaProgram.QuantumElementReference(name=element, loc=loc),
            )
        )
        scopes_manager.append_statement(statement)


def reset_frame(*elements: str) -> None:
    """Resets the frame of the oscillator associated with an element to 0.

    Used to reset all the frame updated made up to this statement.

    Args:
        *elements (str): a single element whose oscillator's phase will
            be reset. multiple elements can be given, in which case all
            of their oscillators' phases will be reset

    """
    loc = _get_loc()
    for element in elements:
        statement = inc_qua_pb2.QuaProgram.AnyStatement(
            resetFrame=inc_qua_pb2.QuaProgram.ResetFrameStatement(
                loc=loc,
                qe=inc_qua_pb2.QuaProgram.QuantumElementReference(name=element, loc=loc),
            )
        )
        scopes_manager.append_statement(statement)


def fast_frame_rotation(cosine: Scalar[float], sine: Scalar[float], *elements: str) -> None:
    r"""Shift the phase of the oscillator associated with an element by applying the
    rotation matrix [[cosine, -sine],[sin, cosine]].

    This is typically used for virtual z-rotations.

    -- Available from QOP 2.2 --

    Note:
        The phase is accumulated with a resolution of 16 bit.
        Therefore, *N* changes to the phase can result in a phase (and amplitude) inaccuracy of about :math:`N \cdot 2^{-16}`.
        To null out this accumulated error, it is recommended to use `reset_frame(el)` from time to time.

    Args:
        cosine (Union[float,QUA variable of type real]): The main
            diagonal values of the rotation matrix
        sine (Union[float,QUA variable of type real]): The bottom left
            rotation matrix element and minus the top right rotation
            matrix element value
        *elements (str): A single element whose oscillator's phase will
            be shifted. multiple elements can be given, in which case
            all of their oscillators' phases will be shifted
    """
    scopes_manager.program_scope.add_used_capability(QopCaps.fast_frame_rotation)
    loc = _get_loc()
    for element in elements:
        statement = inc_qua_pb2.QuaProgram.AnyStatement(
            fastFrameRotation=inc_qua_pb2.QuaProgram.FastFrameRotationStatement(
                loc=loc,
                cosine=to_scalar_pb_expression(cosine),
                sine=to_scalar_pb_expression(sine),
                qe=inc_qua_pb2.QuaProgram.QuantumElementReference(name=element, loc=loc),
            )
        )
        scopes_manager.append_statement(statement)



from enum import IntEnum
from dataclasses import fields, is_dataclass
from typing import Any, Type, Union, get_args, get_origin, get_type_hints

from google.protobuf.message import Message

from qm.grpc.qm.pb import inc_qua_config_pb2
from qm.utils.config_utils import get_controller_pb_config


def fill_defaults_in_config_v1(config: inc_qua_config_pb2.QuaConfig) -> None:
    """
    Since proto values for the controller configuration were made optional, a subtle bug was introduced in the older
    config version (v1). When fetching the config from the gateway, it returns None for fields with default gRPC values.
    For instance, the default gRPC value for a float is 0.0. If a user set an offset of 0.0 and fetched the config,
    they would receive None. This issue arose with the introduction of "optional" labels in config_v2, because beforehand
    when fields were not marked as optional, the None values would automatically be converted to the default grpc value.
    TLDR: SDK version with optional labels result in None values when working with gateway version that doesn't have
    optional labels.
    Consequently, validation tests such as test_config_supports_shareable_ports and test_config_without_shareable would
    fail.
    """
    controller_config = get_controller_pb_config(config)
    if isinstance(controller_config, inc_qua_config_pb2.QuaConfig.QuaConfigV1):
        for _, control_devices_data in controller_config.controlDevices.items():
            _fill_defaults(control_devices_data)

        for _, controllers_data in controller_config.controllers.items():
            _fill_defaults(controllers_data)


# These fields should remain with None values.
FIELDS_TO_EXCLUDE = [
    "high_pass"
]  # The high_pass field was configured as optional before config_v2, and therefore should remain as None.

PRIMITIVE_TYPES = (int, float, str, bool)


def _fill_defaults(obj: Message) -> None:
    """Fill default values for fields of a single message object (with recursive functionality)."""
    if not is_dataclass(obj):
        return

    hints = get_type_hints(type(obj))
    for field in fields(obj):
        field_name = field.name

        # Skip unset "oneof" fields to avoid exceptions when accessing them. Also skip fields that are explicitly excluded.
        if not hasattr(obj, field_name) or field_name in FIELDS_TO_EXCLUDE:
            continue

        value = getattr(obj, field_name)

        typ = _unwrap_optional(hints.get(field_name))

        if value is None:
            default_value = _get_default_value(typ)
            if default_value is not None:
                setattr(obj, field_name, default_value)
        else:
            _apply_to_nested(value)


def _unwrap_optional(typ: Type[Any]) -> Any:
    """Return the inner type from Optional[T], or the original type."""
    origin = get_origin(typ)
    if origin is Union:
        args = get_args(typ)
        none_type = type(None)  # We don't use it directly in the list comprehension just for flake8 reasons
        non_none_args = [arg for arg in args if arg is not none_type]
        if len(non_none_args) > 1:
            raise TypeError(
                f"Unexpected Union with multiple non-None types: {non_none_args}."
                f"Expected a simple Optional[T] (i.e., Union[T, NoneType]) with exactly one non-None type."
            )
        return non_none_args[0]
    return typ


def _apply_to_nested(value: Any) -> None:
    """Recursively apply the 'fill_defaults' function to messages, lists, or dicts of messages."""
    if isinstance(value, PRIMITIVE_TYPES):
        # Not nested, so we skip
        pass
    elif isinstance(value, Message):
        _fill_defaults(value)
    # Tuples and sets aren't standard in proto, but included for forward compatibility and custom extensions.
    elif isinstance(value, list) or isinstance(value, tuple) or isinstance(value, set):
        for item in value:
            _apply_to_nested(item)
    elif isinstance(value, dict):
        for val in value.values():
            _apply_to_nested(val)
    else:
        raise TypeError(f"Unsupported nested value type: {type(value)}")


def _get_default_value(typ: Type[Any]) -> Any:
    """Return the default value for the given type."""
    if typ in PRIMITIVE_TYPES:
        return typ()  # Returns the default value for each primitive type (0 for int, etc.)
    if isinstance(typ, type) and issubclass(typ, IntEnum):
        # For enums, the first value is the default value
        return next(iter(typ))
    return None



from typing import Union

from qm.grpc.qm.pb import inc_qua_config_pb2
from qm.api.models.capabilities import QopCaps
from qm.utils.protobuf_utils import which_one_of, proto_repeated_to_list
from qm.program._validate_config_schema import validate_arbitrary_waveform
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter
from qm.type_hinting.config_types import (
    WaveformArrayConfigType,
    ConstantWaveformConfigType,
    ArbitraryWaveformConfigType,
)

WaveformDictConfigTypes = Union[ConstantWaveformConfigType, ArbitraryWaveformConfigType, WaveformArrayConfigType]


class WaveformConverter(BaseDictToPbConverter[WaveformDictConfigTypes, inc_qua_config_pb2.QuaConfig.WaveformDec]):
    def convert(self, input_data: WaveformDictConfigTypes) -> inc_qua_config_pb2.QuaConfig.WaveformDec:
        if input_data["type"] == "constant":
            return self.constant_waveform_to_protobuf(input_data)
        elif input_data["type"] == "arbitrary":
            return self.arbitrary_waveform_to_protobuf(input_data)
        elif input_data["type"] == "array":
            return self.waveform_array_to_protobuf(input_data)
        else:
            raise ValueError("Unknown waveform type")

    @staticmethod
    def constant_waveform_to_protobuf(data: ConstantWaveformConfigType) -> inc_qua_config_pb2.QuaConfig.WaveformDec:
        return inc_qua_config_pb2.QuaConfig.WaveformDec(
            constant=inc_qua_config_pb2.QuaConfig.ConstantWaveformDec(sample=data["sample"])
        )

    @staticmethod
    def arbitrary_waveform_to_protobuf(data: ArbitraryWaveformConfigType) -> inc_qua_config_pb2.QuaConfig.WaveformDec:
        wf = inc_qua_config_pb2.QuaConfig.WaveformDec()

        is_overridable = data.get("is_overridable", False)
        has_max_allowed_error = "max_allowed_error" in data
        has_sampling_rate = "sampling_rate" in data
        validate_arbitrary_waveform(is_overridable, has_max_allowed_error, has_sampling_rate)

        wf.arbitrary.CopyFrom(
            inc_qua_config_pb2.QuaConfig.ArbitraryWaveformDec(samples=data["samples"], isOverridable=is_overridable)
        )

        if has_max_allowed_error:
            wf.arbitrary.maxAllowedError.value = data["max_allowed_error"]
        elif has_sampling_rate:
            wf.arbitrary.samplingRate.value = data["sampling_rate"]
        elif not is_overridable:
            wf.arbitrary.maxAllowedError.value = 1e-4
        return wf

    def waveform_array_to_protobuf(
        self,
        data: WaveformArrayConfigType,
    ) -> inc_qua_config_pb2.QuaConfig.WaveformDec:
        self._capabilities.validate({QopCaps.waveform_array})

        return inc_qua_config_pb2.QuaConfig.WaveformDec(
            array=inc_qua_config_pb2.QuaConfig.WaveformArrayDec(
                samples_array=[
                    inc_qua_config_pb2.QuaConfig.WaveformSamples(samples=list(samples))
                    for samples in data["samples_array"]
                ]
            )
        )

    def deconvert(self, output_data: inc_qua_config_pb2.QuaConfig.WaveformDec) -> WaveformDictConfigTypes:
        _, curr_waveform = which_one_of(output_data, "waveform_oneof")
        if isinstance(curr_waveform, inc_qua_config_pb2.QuaConfig.ArbitraryWaveformDec):
            arbitrary_waveform_dict: ArbitraryWaveformConfigType = {
                "type": "arbitrary",
                "samples": proto_repeated_to_list(curr_waveform.samples),
                "is_overridable": curr_waveform.isOverridable,
            }
            if curr_waveform.HasField("maxAllowedError"):
                arbitrary_waveform_dict["max_allowed_error"] = curr_waveform.maxAllowedError.value
            if curr_waveform.HasField("samplingRate"):
                arbitrary_waveform_dict["sampling_rate"] = curr_waveform.samplingRate.value
            return arbitrary_waveform_dict

        elif isinstance(curr_waveform, inc_qua_config_pb2.QuaConfig.ConstantWaveformDec):
            constant_waveform_dict: ConstantWaveformConfigType = {
                "type": "constant",
                "sample": curr_waveform.sample,
            }
            return constant_waveform_dict

        elif isinstance(curr_waveform, inc_qua_config_pb2.QuaConfig.WaveformArrayDec):
            waveform_array_dict: WaveformArrayConfigType = {
                "type": "array",
                "samples_array": [waveform_samples.samples for waveform_samples in curr_waveform.samples_array],
            }
            return waveform_array_dict

        else:
            raise Exception(f"Unknown waveform type - {type(curr_waveform)}")



from typing import Dict, Type, Tuple, Union, cast

from qm.exceptions import QmQuaException
from qm.grpc.qm.pb import errors_pb2, job_manager_pb2

ResponseType = Union[
    job_manager_pb2.InsertInputStreamResponse,
    job_manager_pb2.SetElementCorrectionResponse,
    job_manager_pb2.GetElementCorrectionResponse,
]

RequestType = Union[
    job_manager_pb2.InsertInputStreamRequest,
    job_manager_pb2.SetElementCorrectionRequest,
    job_manager_pb2.GetElementCorrectionRequest,
]


class QmApiError(QmQuaException):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "QmApiError":
        return QmApiError(0, str(response))

    def __init__(self, code: int, message: str = "") -> None:
        super().__init__(message)
        self.message = message
        self.code = code


class UnspecifiedError(QmApiError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "UnspecifiedError":
        return UnspecifiedError("Unspecified operation specific error")

    def __init__(self, message: str) -> None:
        super().__init__(0, message)


class _QmJobError(QmApiError):
    """Base class for exceptions in this module."""

    pass


class MissingJobError(_QmJobError):
    """if the job isn't recognized by the server (can happen if it never ran or if it was already deleted)"""

    def __init__(self) -> None:
        super().__init__(1000)


class InvalidJobExecutionStatusError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1001)


class InvalidOperationOnSimulatorJobError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1002)


class InvalidOperationOnRealJobError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1003)


class UnknownInputStreamError(_QmJobError):
    def __init__(self) -> None:
        super().__init__(1006)


class ConfigQueryError(QmApiError):
    pass


class MissingElementError(ConfigQueryError):
    """"""

    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "MissingElementError":
        return MissingElementError(response.jobManagerResponseHeader.jobErrorDetails.message)

    def __init__(self, message: str):
        super().__init__(4001, message)


class MissingDigitalInputError(ConfigQueryError):
    """"""

    @staticmethod
    def _build_from_response(request: RequestType, response: ResponseType) -> "MissingDigitalInputError":
        return MissingDigitalInputError(response.jobManagerResponseHeader.jobErrorDetails.message)

    def __init__(self, message: str):
        super().__init__(4002, message)


class _InvalidConfigChangeError(QmApiError):
    pass


class ElementWithSingleInputError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "ElementWithSingleInputError":
        return ElementWithSingleInputError(
            cast(
                Union[job_manager_pb2.SetElementCorrectionRequest, job_manager_pb2.GetElementCorrectionRequest], request
            ).qeName
        )

    def __init__(self, element_name: str):
        super().__init__(3000)
        self.element_name = element_name


class InvalidElementCorrectionError(_InvalidConfigChangeError):
    """If the correction values are invalid"""

    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "InvalidElementCorrectionError":
        return InvalidElementCorrectionError(
            response.jobManagerResponseHeader.jobErrorDetails.message,
            cast(
                Union[job_manager_pb2.SetElementCorrectionRequest, job_manager_pb2.GetElementCorrectionRequest], request
            ).qeName,
            (
                cast(
                    Union[job_manager_pb2.SetElementCorrectionRequest, job_manager_pb2.GetElementCorrectionRequest],
                    request,
                ).correction.v00,
                cast(
                    Union[job_manager_pb2.SetElementCorrectionRequest, job_manager_pb2.GetElementCorrectionRequest],
                    request,
                ).correction.v01,
                cast(
                    Union[job_manager_pb2.SetElementCorrectionRequest, job_manager_pb2.GetElementCorrectionRequest],
                    request,
                ).correction.v10,
                cast(
                    Union[job_manager_pb2.SetElementCorrectionRequest, job_manager_pb2.GetElementCorrectionRequest],
                    request,
                ).correction.v11,
            ),
        )

    def __init__(
        self,
        message: str,
        element_name: str,
        correction: Tuple[float, float, float, float],
    ) -> None:
        super().__init__(3001, message)
        self.element_name = element_name
        self.correction = correction


class ElementWithoutIntermediateFrequencyError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "ElementWithoutIntermediateFrequencyError":
        return ElementWithoutIntermediateFrequencyError(
            cast(
                Union[job_manager_pb2.SetElementCorrectionRequest, job_manager_pb2.GetElementCorrectionRequest], request
            ).qeName
        )

    def __init__(self, element_name: str):
        super().__init__(3002)
        self.element_name = element_name


class InvalidDigitalInputThresholdError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "InvalidDigitalInputThresholdError":
        return InvalidDigitalInputThresholdError(response.jobManagerResponseHeader.jobErrorDetails.message)

    def __init__(self, message: str):
        super().__init__(3003)
        self.message = message


class InvalidDigitalInputDeadtimeError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "InvalidDigitalInputDeadtimeError":
        return InvalidDigitalInputDeadtimeError(response.jobManagerResponseHeader.jobErrorDetails.message)

    def __init__(self, message: str):
        super().__init__(3004)
        self.message = message


class InvalidDigitalInputPolarityError(_InvalidConfigChangeError):
    @staticmethod
    def build_from_response(request: RequestType, response: ResponseType) -> "InvalidDigitalInputPolarityError":
        return InvalidDigitalInputPolarityError(response.jobManagerResponseHeader.jobErrorDetails.message)

    def __init__(self, message: str):
        super().__init__(3005)
        self.message = message


def _handle_job_manager_error(
    request: RequestType,
    response: ResponseType,
    valid_errors: Tuple[Type[QmQuaException], ...],
) -> None:
    api_response: job_manager_pb2.JobManagerResponseHeader = response.jobManagerResponseHeader
    if not api_response.success:
        error_type = api_response.jobManagerErrorType

        if error_type == errors_pb2.JobManagerErrorTypes.MissingJobError:
            raise MissingJobError()
        elif error_type == errors_pb2.JobManagerErrorTypes.InvalidJobExecutionStatusError:
            raise InvalidJobExecutionStatusError()
        elif error_type == errors_pb2.JobManagerErrorTypes.InvalidOperationOnSimulatorJobError:
            raise InvalidOperationOnSimulatorJobError()
        elif error_type == errors_pb2.JobManagerErrorTypes.InvalidOperationOnRealJobError:
            raise InvalidOperationOnRealJobError()
        elif error_type == errors_pb2.JobManagerErrorTypes.JobOperationSpecificError:
            exception_to_raise = _get_handle_job_operation_error(request, response)
            if exception_to_raise is not None and type(exception_to_raise) in valid_errors:
                raise exception_to_raise
            else:
                raise UnspecifiedError("Unspecified operation specific error")
        elif error_type == errors_pb2.JobManagerErrorTypes.ConfigQueryError:
            exception_to_raise = _get_handle_config_query_error(request, response)
            if exception_to_raise is not None and type(exception_to_raise) in valid_errors:
                raise exception_to_raise
            else:
                raise UnspecifiedError("Unspecified operation specific error")
        elif error_type == errors_pb2.JobManagerErrorTypes.UnknownInputStreamError:
            raise UnknownInputStreamError()
        else:
            raise UnspecifiedError("Unspecified operation error")


def _get_handle_config_query_error(request: RequestType, response: ResponseType) -> QmApiError:
    error_type = response.jobManagerResponseHeader.jobErrorDetails.configQueryErrorType
    errors: Dict[int, Type[QmApiError]] = {
        errors_pb2.ConfigQueryErrorTypes.MissingElementError: MissingElementError,
        errors_pb2.ConfigQueryErrorTypes.MissingDigitalInputError: MissingDigitalInputError,
    }

    return errors.get(error_type, UnspecifiedError).build_from_response(request, response)


def _get_handle_job_operation_error(request: RequestType, response: ResponseType) -> QmApiError:
    error_type = response.jobManagerResponseHeader.jobErrorDetails.jobOperationSpecificErrorType
    errors: Dict[int, Type[QmApiError]] = {
        errors_pb2.JobOperationSpecificErrorTypes.SingleInputElementError: ElementWithSingleInputError,
        errors_pb2.JobOperationSpecificErrorTypes.InvalidCorrectionMatrixError: InvalidElementCorrectionError,
        errors_pb2.JobOperationSpecificErrorTypes.ElementWithoutIntermediateFrequencyError: ElementWithoutIntermediateFrequencyError,
        errors_pb2.JobOperationSpecificErrorTypes.InvalidDigitalInputThresholdError: InvalidDigitalInputThresholdError,
        errors_pb2.JobOperationSpecificErrorTypes.InvalidDigitalInputDeadtimeError: InvalidDigitalInputDeadtimeError,
        errors_pb2.JobOperationSpecificErrorTypes.InvalidDigitalInputPolarityError: InvalidDigitalInputPolarityError,
    }

    return errors.get(error_type, UnspecifiedError).build_from_response(request, response)



from typing import Iterable
from typing import Optional
from typing import Iterable as IterableClass
from typing import Tuple, Union, Literal, Mapping

from qm._loc import _get_loc
from qm.type_hinting import NumberT
from qm.grpc.qm.pb import inc_qua_pb2
from qm.exceptions import QmQuaException
from qm.qua._dsl.variable_handling import declare
from qm.qua._dsl.measure.measure import MeasurePulseType
from qm.qua._dsl._utils import _standardize_timestamp_label
from qm.qua._scope_management.scopes_manager import scopes_manager
from qm.qua._dsl.stream_processing.stream_processing import StreamType
from qm.qua._expressions import Scalar, QuaArrayVariable, to_scalar_pb_expression, create_qua_scalar_expression

PlayPulseType = Union[MeasurePulseType, inc_qua_pb2.QuaProgram.RampPulse]

ChirpUnits = Literal[
    "Hz/nsec",
    "GHz/sec",
    "mHz/nsec",
    "MHz/sec",
    "uHz/nsec",
    "KHz/sec",
    "nHz/nsec",
    "Hz/sec",
    "pHz/nsec",
    "mHz/sec",
]
ChirpType = Union[
    Tuple[Union[Iterable[int], QuaArrayVariable[int], Scalar[int]], ChirpUnits],
    Tuple[Iterable[int], Iterable[int], ChirpUnits],
]
"""A type for performing piecewise linear sweep of the element’s intermediate frequency in time.
A tuple, with the 1st element being a list of rates and the second should be a string with the unit type.
See the ChirpUnits type for the complete list of supported units.
"""


def play(
    pulse: PlayPulseType,
    element: str,
    duration: Optional[Scalar[int]] = None,
    condition: Optional[Scalar[bool]] = None,
    chirp: Optional[ChirpType] = None,
    truncate: Optional[Scalar[int]] = None,
    timestamp_stream: Optional[StreamType] = None,
    continue_chirp: bool = False,
    target: str = "",
) -> None:
    r"""Play a `pulse` based on an 'operation' defined in `element`.

    The pulse will be modified according to the properties of the element
    (see detailed explanation about pulse modifications below),
    and then played to the OPX output(s) defined to be connected
    to the input(s) of the element in the configuration.

    Args:
        pulse (str): The name of an `operation` to be performed, as
            defined in the element in the quantum machine configuration.
            Can also be a [ramp][qm.qua.ramp] function or be multiplied by an
            [ramp][qm.qua.ramp].
        element (str): The name of the element, as defined in the
            quantum machine configuration.
        duration (Union[int,QUA variable of type int]): The time to play
            this pulse in units of the clock cycle (4ns). If not
            provided, the default pulse duration will be used. It is
            possible to dynamically change the duration of both constant
            and arbitrary pulses. Arbitrary pulses can only be stretched,
            not compressed.
        chirp (ChirpType): Allows to perform piecewise linear sweep
            of the element’s intermediate frequency in time. Input
            should be a tuple, with the 1st element being a list of
            rates and the second should be a string with the units.
            The units can be either: ‘Hz/nsec’, ’mHz/nsec’, ’uHz/nsec’,
            ’pHz/nsec’ or ‘GHz/sec’, ’MHz/sec’, ’KHz/sec’, ’Hz/sec’, ’mHz/sec’.
        truncate (Union[int, QUA variable of type int]): Allows playing
            only part of the pulse, truncating the end. If provided,
            will play only up to the given time in units of the clock
            cycle (4ns).
        condition (A logical expression to evaluate.): Will play the operation only if the condition is true.
            The play command will take the same amount of time regardless of the condition (If false, would wait instead).
            Prior to QOP 2.2, only the analog part was conditioned, i.e., any digital pulses associated
            with the operation would always play.
        timestamp_stream (Union[str, ResultStreamSource]): (Supported from
            QOP 2.2) Adding a `timestamp_stream` argument will save the
            time at which the operation occurred to a stream. If the
            `timestamp_stream` is a string ``label``, then the timestamp
            handle can be retrieved with
            [`JobResults.get`][qm.StreamsManager.get] with the same
            ``label``.
        continue_chirp (bool): When performing a chirp, passing `True` will make the chirp continue until a new chirp command is given. Defaults to `False`. Not available in OPX1.0.
        target (str): Allows to select a specific input of the element to play the pulse on. Only allowed (and required) when the element is defined with `singleInputCollection`.

    Note:
        Arbitrary waveforms cannot be compressed and can only be expanded up to
        $2^{24}-1$ clock cycles (67ms). Unexpected output will occur if a duration
        outside the range is given.
        See [Dynamic pulse duration](../../Guides/features.md#dynamic-pulse-duration)
        in the documentation for further information.

    Note:
        When using chirp, it is possible to add a flag "continue_chirp=True" to the play command.
        When this flag is set, the internal oscillator will continue the chirp even after the play command had ended.
        See the `chirp documentation [chirp documentation](../../Guides/features.md#frequency-chirp)
        for more information.

    Example:
        ```python
        with program() as prog:
            v1 = declare(fixed)
            assign(v1, 0.3)
            play('pulse1', 'element1')
            play('pulse1' * amp(0.5), 'element1')
            play('pulse1' * amp(v1), 'element1')
            play('pulse1' * amp(0.9, v1, -v1, 0.9), 'element_iq_pair')
            time_stream = declare_output_stream()
            # Supported on QOP2.2+
            play('pulse1', 'element1', duration=16, timestamp_stream='t1')
            play('pulse1', 'element1', duration=16, timestamp_stream=time_stream)
            with stream_processing():
                stream.buffer(10).save_all('t2')
        ```
    """
    condition_ = to_scalar_pb_expression(condition) if condition is not None else None
    duration_ = to_scalar_pb_expression(duration) if duration is not None else None
    chirp_ = _standardize_chirp(chirp, continue_chirp)
    truncate_ = to_scalar_pb_expression(truncate) if truncate is not None else None
    timestamp_label = _standardize_timestamp_label(timestamp_stream)

    amp = None
    if isinstance(pulse, tuple):
        pulse, amp = pulse

    loc = _get_loc()
    play_statement = inc_qua_pb2.QuaProgram.PlayStatement(
        loc=loc,
        qe=inc_qua_pb2.QuaProgram.QuantumElementReference(name=element, loc=loc),
        targetInput=target,
    )
    if isinstance(pulse, inc_qua_pb2.QuaProgram.RampPulse):
        play_statement.rampPulse.CopyFrom(pulse)
    else:
        play_statement.namedPulse.CopyFrom(inc_qua_pb2.QuaProgram.PulseReference(name=pulse, loc=loc))

    if duration_ is not None:
        play_statement.duration.CopyFrom(duration_)
    if condition_ is not None:
        play_statement.condition.CopyFrom(condition_)
    if chirp_ is not None:
        play_statement.chirp.CopyFrom(chirp_)
        play_statement.chirp.loc = loc
    if amp is not None:
        play_statement.amp.CopyFrom(inc_qua_pb2.QuaProgram.AmpMultiplier(loc=loc, v0=amp[0]))
        for i in range(1, 4, 1):
            if amp[i] is not None:
                getattr(play_statement.amp, "v" + str(i)).CopyFrom(amp[i])
    if truncate_ is not None:
        play_statement.truncate.CopyFrom(truncate_)
    if timestamp_label is not None:
        play_statement.timestampLabel = timestamp_label
    _port_condition = scopes_manager.port_condition
    if _port_condition is not None:
        play_statement.port_condition.CopyFrom(_port_condition)

    scopes_manager.append_statement(inc_qua_pb2.QuaProgram.AnyStatement(play=play_statement))


def _standardize_chirp(chirp: Optional[ChirpType], continue_chirp: bool) -> Optional[inc_qua_pb2.QuaProgram.Chirp]:
    if chirp is None:
        return None

    if len(chirp) == 2:
        chirp_var, chirp_units = chirp
        chirp_times = None
    elif len(chirp) == 3:
        chirp_var, chirp_times, chirp_units = chirp
    else:
        raise QmQuaException("chirp must be tuple of 2 or 3 values")
    chirp_times_list = [int(x) for x in chirp_times] if chirp_times is not None else None
    if isinstance(chirp_var, IterableClass):
        chirp_var = [int(x) for x in chirp_var]
        chirp_var = declare(int, value=chirp_var)

    chirp_obj = inc_qua_pb2.QuaProgram.Chirp()
    chirp_obj.continueChirp = continue_chirp
    if chirp_times_list is not None:
        chirp_obj.times.extend(chirp_times_list)
    if isinstance(chirp_var, QuaArrayVariable):
        chirp_obj.arrayRate.CopyFrom(chirp_var.unwrapped)
    else:
        chirp_obj.scalarRate.CopyFrom(to_scalar_pb_expression(chirp_var))

    units_mapping: Mapping[ChirpUnits, int] = {
        "Hz/nsec": inc_qua_pb2.QuaProgram.Chirp.Units.HzPerNanoSec,
        "GHz/sec": inc_qua_pb2.QuaProgram.Chirp.Units.HzPerNanoSec,
        "mHz/nsec": inc_qua_pb2.QuaProgram.Chirp.Units.mHzPerNanoSec,
        "MHz/sec": inc_qua_pb2.QuaProgram.Chirp.Units.mHzPerNanoSec,
        "uHz/nsec": inc_qua_pb2.QuaProgram.Chirp.Units.uHzPerNanoSec,
        "KHz/sec": inc_qua_pb2.QuaProgram.Chirp.Units.uHzPerNanoSec,
        "nHz/nsec": inc_qua_pb2.QuaProgram.Chirp.Units.nHzPerNanoSec,
        "Hz/sec": inc_qua_pb2.QuaProgram.Chirp.Units.nHzPerNanoSec,
        "pHz/nsec": inc_qua_pb2.QuaProgram.Chirp.Units.pHzPerNanoSec,
        "mHz/sec": inc_qua_pb2.QuaProgram.Chirp.Units.pHzPerNanoSec,
    }

    if chirp_units in units_mapping:
        chirp_obj.units = units_mapping[chirp_units]  # type: ignore[assignment]
    else:
        raise QmQuaException(f'unknown units "{chirp_units}"')
    return chirp_obj


def ramp(v: Scalar[NumberT]) -> inc_qua_pb2.QuaProgram.RampPulse:
    """To be used only within a [`play`][qm.qua.play] command, instead of the `operation`.

    It’s possible to generate a voltage ramp by using the `ramp(slope)` command.
    The slope argument is specified in units of `V/ns`. Usage of this feature is as follows:

    ``play(ramp(0.0001), 'qe1', duration=1000)``

    note:
        The pulse duration must be specified if the ramp feature is used.

    Args:
        v: The slope in units of `V/ns`
    """
    value = create_qua_scalar_expression(v)
    result = inc_qua_pb2.QuaProgram.RampPulse(value=value.unwrapped)
    return result



from typing import Optional, overload

from qm.type_hinting import NumberT
from qm.exceptions import QmQuaException
from qm.qua._dsl.variable_handling import advance_input_stream
from qm.qua._dsl.streams.external_streams import _receive_from_opnic_stream
from qm.qua._expressions import (
    StructT,
    QuaArrayInputStream,
    InputStreamInterface,
    QuaVariableInputStream,
    QuaExternalIncomingStream,
)


@overload
def receive_from_stream(stream: QuaVariableInputStream[NumberT]) -> None:
    """
    Receive the next scalar value from a client input stream.

    This is equivalent to [qm.qua.advance_input_stream][].

    Args:
        stream (QuaVariableInputStream[NumberT]): The client input stream to advance.
    """
    ...


@overload
def receive_from_stream(stream: QuaArrayInputStream[NumberT]) -> None:
    """
    Receive the next array from a client input stream.

    This is equivalent to [qm.qua.advance_input_stream][].

    Args:
        stream (QuaArrayInputStream[NumberT]): The client input stream to advance.
    """
    ...


@overload
def receive_from_stream(stream: QuaExternalIncomingStream[StructT], *, target_variable: StructT) -> None:
    """
    Receive the next packet from an incoming OPNIC stream.

    Args:
        stream (QuaExternalIncomingStream[StructT]): The incoming OPNIC stream to receive from.
        target_variable (StructT): The struct instance that will hold the received packet.
            This must be declared with [qm.qua.declare_struct][].
    """
    ...


def receive_from_stream(stream: InputStreamInterface, *, target_variable: Optional[StructT] = None) -> None:
    """
    Receive the next item from an input stream.

    This is a blocking operation for both client and OPNIC input streams.
    If no data is available yet, the OPX waits until the next value or packet arrives.

    For client input streams, this is equivalent to [qm.qua.advance_input_stream][]:
    it advances the stream so the declared QUA variable or array now holds the next queued value.

    For OPNIC input streams, this receives the next packet and copies it into ``target_variable``.

    Args:
        stream: The incoming stream to receive data from.
        target_variable: A struct instance declared with [qm.qua.declare_struct][] that will hold
            the received packet. This argument is required for OPNIC streams and must not be passed
            for client streams.

    Example:
        ```python
        tau = declare_input_stream("client", "tau", int)
        receive_from_stream(tau)  # This advances tau to the next queued value.

        # OPNIC usage example
        packet_stream = declare_input_stream("opnic", 1, Packet)
        packet = declare_struct(Packet)
        receive_from_stream(packet_stream, target_variable=packet)  # This updates packet in place.
        ```
    """
    if isinstance(stream, QuaVariableInputStream) or isinstance(stream, QuaArrayInputStream):
        if target_variable is not None:
            raise QmQuaException("Client input streams can not receive data into a target variable.")
        advance_input_stream(stream)

    elif isinstance(stream, QuaExternalIncomingStream):
        if target_variable is None:
            raise QmQuaException("'target_variable' must be provided when receiving data from an opnic stream.")
        # Validation of target_variable type is done in the inner '_receive_from_opnic_stream' function.
        _receive_from_opnic_stream(stream, target_variable)

    else:
        raise QmQuaException(f"Unsupported stream type: {type(stream).__name__}.")



import re
import logging
from typing import Any, Optional
from collections.abc import Mapping, Callable, MutableSequence

from google.protobuf.message import Message
from google.protobuf.struct_pb2 import Value, ListValue

from qm.exceptions import QmQuaException
from qm.grpc.qm.pb.inc_qua_config_pb2 import QuaConfig
from qm.serialization.qua_node_visitor import QuaNodeVisitor
from qm.grpc.qm.pb.inc_qua_pb2 import QuaProgram, QuaResultAnalysis
from qm.utils.protobuf_utils import Node, which_one_of, serialized_on_wire
from qm.serialization.expression_serializing_visitor import ExpressionSerializingVisitor

logger = logging.getLogger(__name__)


class InvalidIdentifierError(Exception):
    pass


def _safe_str(value: str) -> str:
    """Safely escape strings for code generation to prevent injection attacks.

    Uses repr() to ensure the string is properly escaped when embedded in generated code.
    This prevents code injection through malicious strings in protobuf fields.

    Args:
        value: The string to escape

    Returns:
        A properly escaped string literal safe for code generation

    Example:
        >>> _safe_str('hello')
        "'hello'"
        >>> _safe_str('x"); os.system("evil')
        '\'x"); os.system("evil\''
    """
    return repr(value)


def _safe_identifier(value: str) -> str:
    """Validate that a string is a safe Python identifier.

    Ensures the value can be used as a variable name without injection risks.

    Args:
        value: The string to validate as an identifier

    Returns:
        The validated identifier

    Raises:
        InvalidIdentifierError: If the value is not a valid Python identifier

    Example:
        >>> _safe_identifier('my_var')
        'my_var'
        >>> _safe_identifier('123invalid')
        InvalidIdentifierError: Invalid Python identifier for a variable name: '123invalid'
    """
    if not value.isidentifier():
        raise InvalidIdentifierError(f"Invalid Python identifier for a variable name: {repr(value)}")
    return value


class StructDeclaration:
    def __init__(self, type_name: str, variable_name: str) -> None:
        self._name: str = type_name
        self.variable_name: str = variable_name
        self._variables: dict[str, tuple[str, int]] = {}

    def variables(self) -> list[str]:
        return list(self._variables.keys())

    def add_variable(self, name: str, var_type: QuaProgram.Type, size: int) -> None:
        self._variables[name] = (_var_type_dec[var_type], size)

    def get_struct_declaration(self) -> str:
        return f"    {self.variable_name} = declare_struct({self._name})"

    def __str__(self) -> str:
        output = f"@qua_struct\nclass {self._name}:\n"
        for name, [var_type, size] in self._variables.items():
            output += f"    {name}: QuaArray[{var_type}, Literal[{size}]]\n"
        output += "\n"
        return output


class QuaSerializingVisitor(QuaNodeVisitor):
    def __init__(self) -> None:
        super().__init__()
        self._indent = 0
        self.structs: list[StructDeclaration] = []
        self._lines: list[str] = []
        self.tags: list[str] = []
        self._used_global_vars: bool = False

    def _format_imports(self) -> str:
        output = ""

        if self.structs:
            output += "\nfrom typing import Literal\n"

        output += """
from qm import CompilerOptionArguments
from qm.qua import *
"""
        if self._used_global_vars:
            output += "from qm.qua._dsl.global_var import assign_global_var, global_var_read, global_var_xor\n"

        output += "\n"
        return output

    def out(self) -> str:
        output = self._format_imports()

        struct_declarations = []
        for struct in self.structs:
            output += str(struct)
            if struct.variable_name:
                struct_declarations.append(struct.get_struct_declaration())

        self._lines = self._lines[0:1] + struct_declarations + self._lines[1:]
        output += "\n".join(self._lines)
        return output

    def _out_lines(self) -> list[str]:
        return self._lines

    def _default_enter(self, node: Node) -> bool:
        if not isinstance(node, tuple(_dont_print)):
            logger.info(f"entering {type(node).__module__}.{type(node).__name__}")
        statement: Optional[Callable[[Node, "QuaSerializingVisitor"], str]] = _statements.get(type(node), None)  # type: ignore[assignment]

        if statement is not None:
            line = statement(node, self)
            self._line(line)

        block: Optional[Callable[[Node, "QuaSerializingVisitor"], str]] = _blocks.get(type(node), None)  # type: ignore[assignment]
        if block is not None:
            self._enter_block(block(node, self))
        return statement is None

    @staticmethod
    def _search_auto_added_stream(values: MutableSequence[Value]) -> bool:
        v2 = values[2]
        is_auto_added_result = False
        if isinstance(v2, Value):
            name, value = which_one_of(v2, "kind")
            if isinstance(value, str):
                is_auto_added_result = name == "string_value" and value == "auto"

        return is_auto_added_result

    def _fix_legacy_save(self, sp_line: str, node: ListValue) -> None:
        split_sp_line = sp_line.split(".")
        trace_name = split_sp_line[0]
        save_name = split_sp_line[-1].split('"')[1]
        save_name = re.sub(r"_input\d", "", save_name)
        is_auto_added_stream = self._search_auto_added_stream(node.values)
        if is_auto_added_stream:
            self._lines.pop()
            line_to_remove_index = None
            for i in range(len(self._lines)):
                if self._lines[i].find(f"{trace_name} = declare_output_stream") > 0:
                    line_to_remove_index = i
                self._lines[i] = re.sub(r"(?<=\W|\^)" + trace_name + r"(?=\W|\$)", f'"{save_name}"', self._lines[i])
            if line_to_remove_index:
                self._lines.pop(line_to_remove_index)

    @property
    def _node_to_enter(self) -> Mapping[type, Callable[[Any], bool]]:
        return {
            ListValue: self.enter_list_value,
            QuaProgram.CompilerOptions: self.enter_compiler_options,
            QuaResultAnalysis: self.enter_result_analysis,
            QuaProgram.SaveStatement: self.enter_save,
            QuaProgram.VarDeclaration: self.enter_variable_declaration,
            QuaProgram.ExternalStreamDeclaration: self.enter_external_stream_declaration,
            QuaProgram.MeasureStatement: self.enter_measure,
        }

    def enter_compiler_options(self, node: QuaProgram.CompilerOptions) -> bool:
        options_list = []
        if node.strict:
            options_list.append(f"strict={node.strict}")
        if len(node.flags) > 0:
            options_list.append(f"flags={node.flags}")
        options_string = ", ".join(options_list)
        if len(options_string) > 0:
            self._lines.insert(0, f"compiler_options = CompilerOptionArguments({options_string})\n")

        return False

    def enter_result_analysis(self, node: QuaResultAnalysis) -> bool:
        if len(node.model) > 0:
            self._enter_block("with stream_processing():")
            return True
        else:
            return False

    def enter_save(self, node: QuaProgram.SaveStatement) -> bool:
        if node.tag not in self.tags:
            self._line(f"{_safe_identifier(node.tag)} = declare_output_stream()")
            self.tags.append(node.tag)
        save_line = f"save({self.serialize_expression(node.source)}, {_safe_identifier(node.tag)})"
        self._line(save_line)
        return False

    def enter_variable_declaration(self, node: QuaProgram.VarDeclaration) -> bool:
        if self._check_if_has_struct(node):
            return False

        args = self._get_declare_var_args(node)

        if node.isInputStream:
            # Removes the '_input_stream' from the end of the name
            stream_name = node.name[13:]
            self._line(
                f"{_safe_identifier(node.name)} = declare_input_stream({_var_type_dec[node.type]}, {_safe_str(stream_name)}, {_dict_to_python_call(args)})"
            )
        else:
            self._line(
                f"{_safe_identifier(node.name)} = declare({_var_type_dec[node.type]}, {_dict_to_python_call(args)})"
            )

        return False

    def _check_if_has_struct(self, node: QuaProgram.VarDeclaration) -> bool:
        if node.structMember is None or not serialized_on_wire(node.structMember):
            return False

        struct: Optional[StructDeclaration] = None
        for _struct in self.structs:
            if _struct.variable_name == node.structMember.name:
                struct = _struct
                break

        if struct is None:
            struct = StructDeclaration(f"Struct{node.structMember.name}", node.structMember.name)
            self.structs.append(struct)

        struct.add_variable(node.name, node.type, node.size)
        return True

    def enter_external_stream_declaration(self, node: QuaProgram.ExternalStreamDeclaration) -> bool:
        struct_type = self._create_struct_from_expected_types(node)
        if node.direction == QuaProgram.Direction.INCOMING:
            name_prefix = "i"
        else:
            name_prefix = "o"

        var_name = f"s{name_prefix}{node.stream_id}"

        if node.direction == QuaProgram.Direction.INCOMING:
            declare_function = "declare_input_stream"
        else:
            declare_function = "declare_output_stream"

        self._line(f'{var_name} = {declare_function}("opnic", {node.stream_id}, {struct_type})')

        return False

    def _create_struct_from_expected_types(self, node: QuaProgram.ExternalStreamDeclaration) -> str:
        struct_class_name = (
            f"StructStream{self.stream_direction_to_string(node.direction).capitalize()}{node.stream_id}"
        )
        struct = StructDeclaration(struct_class_name, "")

        for index, expected_type in enumerate(node.expectedTypes):
            struct.add_variable(f"variable_{index}", expected_type.type, expected_type.size)
        self.structs.append(struct)
        return struct_class_name

    @staticmethod
    def stream_direction_to_string(stream_direction: QuaProgram.Direction) -> str:
        if stream_direction == QuaProgram.Direction.INCOMING:
            return "INCOMING"
        elif stream_direction == QuaProgram.Direction.OUTGOING:
            return "OUTGOING"
        else:
            raise ValueError(f"Unknown stream direction {stream_direction}")

    def enter_list_value(self, node: ListValue) -> bool:
        line = _stream_processing_terminal_statement(node, self)
        self._line(line)
        self._fix_legacy_save(line, node)
        return False

    @property
    def _node_to_leave(self) -> Mapping[type, Callable[[Any], None]]:
        return {QuaProgram: self.leave_program}

    def leave_program(self, node: QuaProgram) -> None:
        if self._lines[-1].find("with stream_processing():") > 0:
            self._lines.pop()

    def _get_declare_var_args(self, node: QuaProgram.VarDeclaration) -> dict[str, str]:
        size = node.size
        dim = node.dim
        args = {}
        if dim > 0:
            args["size"] = f"{size}"
        if dim > 0 and len(node.value) > 0:
            args["value"] = "[" + ", ".join([self.serialize_expression(it) for it in node.value]) + "]"
        elif len(node.value) == 1:
            args["value"] = self.serialize_expression(node.value[0])
        if "value" in args and "size" in args:
            del args["size"]
        return args

    def enter_measure(self, node: QuaProgram.MeasureStatement) -> bool:
        if node.timestampLabel and node.timestampLabel not in self.tags:
            self._line(f"{_safe_identifier(node.timestampLabel)} = declare_output_stream()")
            self.tags.append(node.timestampLabel)
        if node.streamAs and node.streamAs not in self.tags:
            if node.streamAs.startswith("atr_"):  # Support for legacy adc stream naming
                self._line(f"{_safe_identifier(node.streamAs)} = declare_output_stream(adc_trace=True)")
            else:
                self._line(f"{_safe_identifier(node.streamAs)} = declare_output_stream()")
            self.tags.append(node.streamAs)
        return self._default_enter(node)

    @property
    def _node_to_visit(self) -> Mapping[type, Callable[[Any], None]]:
        return {
            QuaProgram.ForStatement: self.visit_for,
            QuaProgram.ForEachStatement: self.visit_for_each,
            QuaProgram.IfStatement: self.visit_if,
            QuaProgram.GlobalVariableAssignmentStatement: self.visit_global_variable_assignment,
            QuaProgram.PlayStatement: self.visit_play,
        }

    def visit_for(self, node: QuaProgram.ForStatement) -> None:
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")

    def visit_for_each(self, node: QuaProgram.ForEachStatement) -> None:
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")

    def visit_if(self, node: QuaProgram.IfStatement) -> None:
        if len(node.body.statements) > 0:
            super()._default_visit(node.body)
        else:
            self._line("pass")
        elseifs_list = node.elseifs
        for elseif in elseifs_list:
            elseif_block = elseif.body
            condition = elseif.condition
            self._leave_block()
            self._line(f"with elif_({self.serialize_expression(condition)}):")
            self._enter_block()
            if len(elseif_block.statements) > 0:
                super()._default_visit(elseif_block)
            else:
                self._line("pass")

        else_block = getattr(node, "else")
        if len(else_block.statements) > 0:  # Cannot ID else_() with pass
            self._leave_block()
            self._line("with else_():")
            self._enter_block()
            super()._default_visit(else_block)

    def visit_global_variable_assignment(self, node: QuaProgram.GlobalVariableAssignmentStatement) -> None:
        self._used_global_vars = True
        self._line(f"assign_global_var({', '.join([self.serialize_expression(v) for v in node.variables])})")

    def visit_play(self, node: QuaProgram.PlayStatement) -> None:
        pulse_one_of, value = which_one_of(node, "pulseType")
        if pulse_one_of == "namedPulse":
            pulse = _safe_str(node.namedPulse.name)
            # node.pulse - duplicate with namedPulse
        elif pulse_one_of == "rampPulse":
            pulse = f"ramp({self.serialize_expression(node.rampPulse)})"
        else:
            raise QmQuaException(f"Unknown pulse type {pulse_one_of}")

        element = node.qe.name
        amp = ""
        if serialized_on_wire(node.amp):
            v0 = self.serialize_expression(node.amp.v0)
            v1 = self.serialize_expression(node.amp.v1)
            v2 = self.serialize_expression(node.amp.v2)
            v3 = self.serialize_expression(node.amp.v3)
            if v0 != "":
                if v1 != "":
                    amp = f"*amp({v0}, {v1}, {v2}, {v3})"
                else:
                    amp = f"*amp({v0})"
        args = []

        _, duration = which_one_of(node.duration, "expression_oneof")
        if duration is not None:
            args.append(f"duration={self.serialize_expression(duration)}")

        _, condition = which_one_of(node.condition, "expression_oneof")
        if condition is not None:
            args.append(f"condition={self.serialize_expression(condition)}")

        if node.targetInput:
            args.append(f"target={_safe_str(node.targetInput)}")

        if serialized_on_wire(node.chirp):
            rate_one_of, value = which_one_of(node.chirp, "rate")
            if isinstance(value, (QuaProgram.AnyScalarExpression, QuaProgram.ArrayVarRefExpression)):
                rate = self.serialize_expression(value)
            else:
                raise QmQuaException(f"Unknown chirp rate {rate_one_of}")

            unit_mapping = {
                QuaProgram.Chirp.Units.HzPerNanoSec: "Hz/nsec",
                QuaProgram.Chirp.Units.mHzPerNanoSec: "mHz/nsec",
                QuaProgram.Chirp.Units.uHzPerNanoSec: "uHz/nsec",
                QuaProgram.Chirp.Units.nHzPerNanoSec: "nHz/nsec",
                QuaProgram.Chirp.Units.pHzPerNanoSec: "pHz/nsec",
            }
            s_units = node.chirp.units
            if s_units in unit_mapping:
                units = unit_mapping[s_units]
            else:
                raise QmQuaException(f"Unsupported units {s_units}")

            times_builder = []
            for time in node.chirp.times:
                times_builder.append(f"{time}")

            if len(times_builder) > 0:
                times = f'[{", ".join(times_builder)}]'
            else:
                times = "None"

            if node.chirp.continueChirp:
                args.append("continue_chirp=True")

            args.append(f'chirp=({rate},{times},"{units}")')

        _, truncate = which_one_of(node.truncate, "expression_oneof")
        if truncate is not None:
            args.append(f"truncate={self.serialize_expression(truncate)}")

        if node.timestampLabel:
            if node.timestampLabel not in self.tags:
                self._line(f"{_safe_identifier(node.timestampLabel)} = declare_output_stream()")
                self.tags.append(node.timestampLabel)
            args.append(f"timestamp_stream={_safe_identifier(node.timestampLabel)}")

        # TODO maybe make sure no other fields?

        if len(args) > 0:
            args_str = f', {", ".join(args)}'
        else:
            args_str = ""
        indent = ""
        if serialized_on_wire(node.port_condition):
            self._line(f"with port_condition({self.serialize_expression(node.port_condition)}):")
            indent = " " * 4
        self._line(f"{indent}play({pulse}{amp}, {_safe_str(element)}{args_str})")

    def _default_leave(self, node: Message) -> None:
        if isinstance(node, tuple(_blocks)):
            self._leave_block()
        super()._default_leave(node)

    def _enter_block(self, line: Optional[str] = None) -> None:
        if line is not None:
            self._line(line)
        self._indent += 1

    def _leave_block(self) -> None:
        self._indent -= 1

    def _line(self, line: str) -> None:
        self._lines.append((self._indent * "    ") + line)

    @property
    def _expression_visitor(self) -> ExpressionSerializingVisitor:
        return ExpressionSerializingVisitor(self)

    def serialize_expression(self, value: Node) -> str:
        return self._expression_visitor.serialize(value)


def _dict_to_python_call(d: dict[Any, Any]) -> str:
    return ", ".join([f"{k}={v}" for k, v in d.items()])


def _ramp_to_zero_statement(node: QuaProgram.RampToZeroStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    args.append(_safe_str(node.qe.name))
    if node.duration is not None:
        args.append(str(node.duration.value))
    return f'ramp_to_zero({", ".join(args)})'


def _measure_statement(node: QuaProgram.MeasureStatement, visitor: QuaSerializingVisitor) -> str:
    args = []

    amp = ""
    v0 = visitor.serialize_expression(node.amp.v0)
    v1 = visitor.serialize_expression(node.amp.v1)
    v2 = visitor.serialize_expression(node.amp.v2)
    v3 = visitor.serialize_expression(node.amp.v3)
    if v0 != "":
        if v1 != "":
            amp = f"*amp({v0}, {v1}, {v2}, {v3})"
        else:
            amp = f"*amp({v0})"

    args.append(f"{_safe_str(node.pulse.name)}{amp}")
    args.append(_safe_str(node.qe.name))

    if len(node.measureProcesses) > 0:
        for process in node.measureProcesses:
            args.append(visitor.serialize_expression(process))
    if node.timestampLabel:
        args.append(f"timestamp_stream={_safe_identifier(node.timestampLabel)}")
    if node.streamAs:
        args.append(f"adc_stream={_safe_identifier(node.streamAs)}")
    return f'measure({", ".join(args)})'


def _load_waveform_statement(node: QuaProgram.LoadWaveformStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    args.append(f"pulse={_safe_str(node.pulse.name)}")
    args.append(f"waveform_index={visitor.serialize_expression(node.waveform_index)}")
    args.append(f"element={_safe_str(node.qe.name)}")
    return f'load_waveform({", ".join(args)})'


def _wait_statement(node: QuaProgram.WaitStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    _, wait_value = which_one_of(node.time, "expression_oneof")
    if wait_value is not None:
        args.append(f"{visitor.serialize_expression(wait_value)}")
    qes = []
    for qe in node.qe:
        qes.append(_safe_str(qe.name))
    args.append(", ".join(qes))
    return f'wait({", ".join(args)})'


def _align_statement(node: QuaProgram.AlignStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    for qe in node.qe:
        args.append(_safe_str(qe.name))
    return f'align({", ".join(args)})'


def _wait_for_trigger_statement(node: QuaProgram.WaitForTriggerStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    for qe in node.qe:
        args.append(_safe_str(qe.name))
    if node.pulseToPlay.name:
        args.append(_safe_str(node.pulseToPlay.name))
    output = which_one_of(node, "source")[1]
    if isinstance(output, QuaProgram.WaitForTriggerStatement.ElementOutput) and output.element:
        if node.elementOutput.output:
            args.append(
                f"trigger_element=({_safe_str(node.elementOutput.element)}, {_safe_str(node.elementOutput.output)})"
            )
        else:
            args.append(f"trigger_element={_safe_str(node.elementOutput.element)}")
    if which_one_of(node.timeTagTarget, "var_oneof")[0] == "name":
        args.append(f"time_tag_target={_safe_identifier(node.timeTagTarget.name)}")
    return f'wait_for_trigger({", ".join(args)})'


def _frame_rotation_statement(node: QuaProgram.ZRotationStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    args.append(f"{visitor.serialize_expression(node.value)}")
    args.append(_safe_str(node.qe.name))
    return f'frame_rotation_2pi({", ".join(args)})'


def _fast_frame_rotation_statement(node: QuaProgram.FastFrameRotationStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    args.append(f"{visitor.serialize_expression(node.cosine)}")
    args.append(f"{visitor.serialize_expression(node.sine)}")
    args.append(_safe_str(node.qe.name))
    return f'fast_frame_rotation({", ".join(args)})'


def _send_to_external_stream(node: QuaProgram.SendToExternalStreamStatement, visitor: QuaSerializingVisitor) -> str:
    name = f"so{node.stream.stream_id}"  # send is only to outgoing streams
    return f"send_to_stream({name}, {_safe_identifier(node.struct.name)})"


def _receive_from_external_stream(
    node: QuaProgram.ReceiveFromExternalStreamStatement, visitor: QuaSerializingVisitor
) -> str:
    name = f"si{node.stream.stream_id}"  # receive is only to incoming streams
    return f"receive_from_stream({name}, target_variable={_safe_identifier(node.struct.name)})"


def _reset_frame_statement(node: QuaProgram.ResetFrameStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    args.append(_safe_str(node.qe.name))
    return f'reset_frame({", ".join(args)})'


def _update_frequency_statement(node: QuaProgram.UpdateFrequencyStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    args.append(_safe_str(node.qe.name))
    args.append(f"{visitor.serialize_expression(node.value)}")
    args.append(_safe_str(QuaProgram.UpdateFrequencyStatement.Units.Name(node.units)))
    args.append(f"{node.keepPhase}")
    return f'update_frequency({", ".join(args)})'


def _set_dc_offset_statement(node: QuaProgram.SetDcOffsetStatement, visitor: QuaSerializingVisitor) -> str:
    args = []
    args.append(_safe_str(node.qe.name))
    args.append(_safe_str(node.qeInputReference))
    args.append(f"{visitor.serialize_expression(node.offset)}")
    return f'set_dc_offset({", ".join(args)})'


def _advance_input_stream_statement(
    node: QuaProgram.AdvanceInputStreamStatement, visitor: QuaSerializingVisitor
) -> str:
    stream_value = which_one_of(node, "stream_oneof")[1]
    if isinstance(stream_value, QuaProgram.ArrayVarRefExpression) and stream_value.name != "":
        input_stream = visitor.serialize_expression(stream_value)
    elif isinstance(stream_value, QuaProgram.VarRefExpression) and stream_value.name != "":
        input_stream = visitor.serialize_expression(stream_value)
    else:
        raise RuntimeError("unsupported type for pop input stream")
    return f"advance_input_stream({input_stream})"


def _for_block_statement(node: QuaProgram.ForStatement, visitor: QuaSerializingVisitor) -> str:
    condition = visitor.serialize_expression(node.condition)
    if len(node.init.statements) == 0 and len(node.update.statements) == 0:
        if condition == "True":
            return "with infinite_loop_():"
        else:
            return f"with while_({condition}):"
    else:
        if len(node.init.statements) != 1:
            raise Exception("for is not valid")
        if len(node.update.statements) != 1:
            raise Exception("for is not valid")

        init = node.init.statements[0].assign
        update = node.update.statements[0].assign

        if init is None:
            raise Exception("for is not valid")
        if update is None:
            raise Exception("for is not valid")

        return f"with for_({visitor.serialize_expression(init.target)},{visitor.serialize_expression(init.expression)},{condition},{visitor.serialize_expression(update.expression)}):"


def _for_each_block_statement(node: QuaProgram.ForEachStatement, visitor: QuaSerializingVisitor) -> str:
    variables = []
    arrays = []
    for it in node.iterator:
        variables.append(visitor.serialize_expression(it.variable))
        arrays.append(it.array.name)
    return f'with for_each_(({",".join(variables)}),({",".join(arrays)})):'


def _if_block_statement(node: QuaProgram.IfStatement, visitor: QuaSerializingVisitor) -> str:
    condition = visitor.serialize_expression(node.condition)
    if node.unsafe is True:
        unsafe = ", unsafe=True"
    else:
        unsafe = ""
    return f"with if_({condition}{unsafe}):"


def _strict_timing_block_statement(_: Any, __: Any) -> str:
    return "with strict_timing_():"


def _stream_processing_function(array: MutableSequence[Value]) -> str:
    function = array[0].string_value

    if function == "average":
        if len(array) > 1:
            first_element = array[1]
            v = which_one_of(first_element, "kind")[1]
            if isinstance(v, str):
                var = v
            elif isinstance(v, ListValue):
                var = _stream_processing_operator(v.values)
            else:
                raise NotImplementedError
        else:
            var = ""
        return f"average({var})"

    if function == "dot":
        if len(array) == 1:
            return "tuple_dot_product()"
        else:
            vector = _stream_processing_operator(array[1].list_value.values)
            return f"dot_product({vector})"

    if function == "vmult":
        vector = _stream_processing_operator(array[1].list_value.values)
        return f"multiply_by({vector})"

    if function == "smult":
        return f"multiply_by({array[1].string_value})"

    if function == "tmult":
        return "tuple_multiply()"

    if function == "conv":
        if len(array) == 2:
            if array[1].string_value:
                mode = f'"{array[1].string_value}"'
            else:
                mode = ""
            return f"tuple_convolution({mode})"
        else:
            vector = _stream_processing_operator(array[2].list_value.values)
            if array[1].string_value:
                mode = f',"{array[1].string_value}"'
            else:
                mode = ""
            return f"convolution({vector}{mode})"

    if function == "fft":
        return "fft()"

    if function == "booleancast":
        return "boolean_to_int()"

    if function == "demod":
        if len(array[2].list_value.values) > 0:
            cos = _stream_processing_operator(array[2].list_value.values)
        else:
            cos = array[2].string_value
        if len(array[3].list_value.values) > 0:
            sin = _stream_processing_operator(array[3].list_value.values)
        else:
            sin = array[3].string_value
        return f"demod({array[1].string_value},{cos},{sin})"

    print(f"missing function: {function}")
    return "default_function()"


def _stream_processing_operator(array: MutableSequence[Value]) -> str:
    operator = array[0].string_value

    if operator == "@macro_input":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.input{array[1].string_value}()"

    if operator == "@array":
        values = []
        for a in array[1:]:
            value = _stream_processing_statement(a)
            values.append(value)
        return f'[{", ".join(values)}]'

    if operator == "@macro_auto_reshape":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.auto_reshape()"

    if operator == "+":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.add({right})"  # arithmetic stream

    if operator == "-":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.subtract({right})"  # arithmetic stream

    if operator == "/":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.divide({right})"  # arithmetic stream

    if operator == "*":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{left}.multiply({right})"  # arithmetic stream

    if operator == "zip":
        left = _stream_processing_statement(array[1])
        right = _stream_processing_statement(array[2])
        return f"{right}.zip({left})"  # arithmetic stream

    if operator == "take":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.take({array[1].string_value})"

    if operator == "buffer":
        chain = _default_stream_processing_chain(array)
        dims = [dim.string_value for dim in array[1:-1]]
        return f'{chain}.buffer({", ".join(dims)})'

    if operator == "bufferAndSkip":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.buffer_and_skip({array[1].string_value}, {array[2].string_value})"

    if operator == "skip":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.skip({array[1].string_value})"

    if operator == "skipLast":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.skip_last({array[1].string_value})"

    if operator == "histogram":
        chain = _default_stream_processing_chain(array)
        bins = _stream_processing_operator(array[1].list_value.values)
        return f"{chain}.histogram({bins})"

    if operator == "average":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.average()"

    if operator == "real":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.real()"

    if operator == "image":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.image()"

    if operator == "flatten":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.flatten()"

    if operator == "map":
        chain = _default_stream_processing_chain(array)
        return f"{chain}.map(FUNCTIONS.{_stream_processing_function(array[1].list_value.values)})"

    if operator == "@re":
        chain = _default_stream_processing_chain(array)
        timestamp_mode = int(array[1].string_value)
        if timestamp_mode == 0:  # values
            return f"{chain}"
        elif timestamp_mode == 1:  # timestamps
            return f"{chain}.timestamps()"
        elif timestamp_mode == 2:  # ValuesAndTimestamps
            return f"{chain}.with_timestamps()"

    if operator == "@macro_adc_trace":
        chain = _default_stream_processing_chain(array)
        return f"{chain}"

    print(f"missing operator: {operator}")
    chain = _default_stream_processing_chain(array)
    return f"{chain}"


def _default_stream_processing_chain(array: MutableSequence[Value]) -> str:
    last_index = len(array) - 1
    chain = _stream_processing_statement(array[last_index])
    return f"{chain}"


def _stream_processing_statement(node: Value) -> str:
    _value = which_one_of(node, "kind")[1]
    if isinstance(_value, ListValue) and len(_value.values) > 0:
        return _stream_processing_operator(_value.values)
    elif isinstance(_value, str):
        return _value
    else:
        raise NotImplementedError(f"Unsupported stream processing statement: {_value}")


def _stream_processing_terminal_statement(node: ListValue, visitor: QuaSerializingVisitor) -> str:
    last_index = len(node.values) - 1
    chain = _stream_processing_statement(node.values[last_index])
    terminal = node.values[0].string_value
    terminal = "save_all" if terminal == "saveAll" else terminal  # normalize save all
    return f'{chain}.{terminal}("{node.values[1].string_value}")'


def _update_correction_statement(node: QuaProgram.UpdateCorrectionStatement, visitor: QuaSerializingVisitor) -> str:
    return (
        f"update_correction({_safe_str(node.qe.name)},{visitor.serialize_expression(node.correction.c0)},"
        f"{visitor.serialize_expression(node.correction.c1)},"
        f"{visitor.serialize_expression(node.correction.c2)},"
        f"{visitor.serialize_expression(node.correction.c3)})"
    )


def _assignment_statement(node: QuaProgram.AssignmentStatement, visitor: QuaSerializingVisitor) -> str:
    return f"assign({visitor.serialize_expression(node.target)}, " f"{visitor.serialize_expression(node.expression)})"


def _serialize(node: Node) -> list[str]:
    visitor = QuaSerializingVisitor()
    visitor.visit(node)
    return visitor._out_lines()


_blocks = {
    QuaProgram: lambda n, v: "with program() as prog:",
    QuaProgram.ForStatement: _for_block_statement,
    QuaProgram.ForEachStatement: _for_each_block_statement,
    QuaProgram.IfStatement: _if_block_statement,
    QuaProgram.StrictTimingStatement: _strict_timing_block_statement,
}


_statements = {
    QuaProgram.MeasureStatement: _measure_statement,
    QuaProgram.WaitStatement: _wait_statement,
    QuaProgram.AssignmentStatement: _assignment_statement,
    QuaProgram.PauseStatement: lambda n, v: "pause()",
    QuaProgram.ResetPhaseStatement: lambda n, v: f"reset_if_phase({_safe_str(n.qe.name)})",
    QuaProgram.ResetGlobalPhaseStatement: lambda n, v: "reset_global_phase()",
    QuaProgram.UpdateFrequencyStatement: _update_frequency_statement,
    QuaProgram.AlignStatement: _align_statement,
    QuaProgram.WaitForTriggerStatement: _wait_for_trigger_statement,
    QuaProgram.ZRotationStatement: _frame_rotation_statement,
    QuaProgram.RampToZeroStatement: _ramp_to_zero_statement,
    QuaProgram.ResetFrameStatement: _reset_frame_statement,
    ListValue: _stream_processing_terminal_statement,
    # list value statement is assumed just as stream processing for now
    QuaProgram.UpdateCorrectionStatement: _update_correction_statement,
    QuaProgram.SetDcOffsetStatement: _set_dc_offset_statement,
    QuaProgram.AdvanceInputStreamStatement: _advance_input_stream_statement,
    QuaProgram.FastFrameRotationStatement: _fast_frame_rotation_statement,
    QuaProgram.LoadWaveformStatement: _load_waveform_statement,
    QuaProgram.SendToExternalStreamStatement: _send_to_external_stream,
    QuaProgram.ReceiveFromExternalStreamStatement: _receive_from_external_stream,
}

_var_type_dec = {
    QuaProgram.Type.INT: "int",
    QuaProgram.Type.BOOL: "bool",
    QuaProgram.Type.REAL: "fixed",
}

_nodes_to_ignore = {
    QuaProgram.Script,
    QuaProgram.StatementsCollection,
    QuaProgram.AnyStatement,
    QuaProgram.AnyScalarExpression,
    QuaProgram.BinaryExpression,
    QuaProgram.VarRefExpression,
    QuaProgram.LiteralExpression,
    QuaProgram.PlayStatement,
    QuaConfig,
}

_dont_print = set(_blocks) | set(_statements) | _nodes_to_ignore
